# -*- coding: utf-8 -*-

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC

from plotpy import io
from plotpy.config import _
from plotpy.constants import X_BOTTOM, Y_LEFT
from plotpy.coords import axes_to_canvas
from plotpy.interfaces import (
    IBaseImageItem,
    IBasePlotItem,
    IColormapImageItemType,
    ICSImageItemType,
    IExportROIImageItemType,
    IHistDataSource,
    IImageItemType,
    ITrackableItemType,
    IVoiImageItemType,
)
from plotpy.items.image.base import BaseImageItem, RawImageItem
from plotpy.items.image.transform import TrImageItem
from plotpy.mathutils.arrayfuncs import get_nan_range
from plotpy.styles import Histogram2DParam, ImageParam, QuadGridParam

try:
    from plotpy._scaler import _histogram, _scale_quads
    from plotpy.histogram2d import histogram2d, histogram2d_func
except ImportError:
    print(
        ("Module 'plotpy.items.image.base': missing C extension"),
        file=sys.stderr,
    )
    print(
        ("try running: python setup.py build_ext --inplace"),
        file=sys.stderr,
    )
    raise

if TYPE_CHECKING:
    import numpy
    import qwt.plot
    import qwt.scale_map
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QPainter

    from plotpy.interfaces import IItemType
    from plotpy.items import RectangleShape
    from plotpy.plot import BasePlot
    from plotpy.styles.base import ItemParameters


class QuadGridItem(RawImageItem):
    """Quad grid item

    Args:
        X: X coordinates
        Y: Y coordinates
        Z: Z coordinates
        param: Image parameters

    X, Y, Z: A structured grid of quadrilaterals, each quad is defined by
    (X[i], Y[i]), (X[i], Y[i+1]), (X[i+1], Y[i+1]), (X[i+1], Y[i])
    """

    __implements__ = (IBasePlotItem, IBaseImageItem, IHistDataSource, IVoiImageItemType)

    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        Z: np.ndarray,
        param: QuadGridParam | None = None,
    ) -> None:
        assert X is not None
        assert Y is not None
        assert Z is not None
        self.X = X
        self.Y = Y
        assert X.shape == Y.shape
        assert Z.shape == X.shape
        super().__init__(Z, param)
        self.set_data(Z)
        self.grid = 1
        self.interpolate = (0, 0.5, 0.5)
        self.param.update_item(self)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> QuadGridParam:
        """Return instance of the default image param DataSet"""
        return QuadGridParam(_("Quadrilaterals"))

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (
            IImageItemType,
            IVoiImageItemType,
            IColormapImageItemType,
            ITrackableItemType,
        )

    def update_bounds(self) -> None:
        """Update image bounds to fit image shape"""
        xmin = self.X.min()
        xmax = self.X.max()
        ymin = self.Y.min()
        ymax = self.Y.max()
        self.bounds = QC.QRectF(xmin, ymin, xmax - xmin, ymax - ymin)

    def set_data(
        self,
        data: np.ndarray,
        X: np.ndarray | None = None,
        Y: np.ndarray | None = None,
        lut_range: list[float, float] | None = None,
    ) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
            X: X coordinates (Default value = None)
            Y: Y coordinates (Default value = None)
            lut_range: LUT range -- tuple (levelmin, levelmax) (Default value = None)
        """
        if lut_range is not None:
            _min, _max = lut_range
        else:
            _min, _max = get_nan_range(data)

        self.data = data
        self.histogram_cache = None
        if X is not None:
            assert Y is not None
            self.X = X
            self.Y = Y

        self.update_bounds()
        self.update_border()
        self.set_lut_range([_min, _max])

    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        if self.warn_if_non_linear_scale(painter, canvasRect):
            return
        self._offscreen[...] = np.uint32(0)
        dest = _scale_quads(
            self.X,
            self.Y,
            self.data,
            src_rect,
            self._offscreen,
            dst_rect,
            self.lut,
            self.interpolate,
            self.grid,
        )
        qrect = QC.QRectF(QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3]))
        painter.drawImage(qrect, self._image, qrect)
        xl, yt, xr, yb = dest
        self._offscreen[yt:yb, xl:xr] = 0

    def notify_new_offscreen(self) -> None:
        """Notify that the offscreen image has changed"""
        # we always ensure the offscreen is clean before drawing
        self._offscreen[...] = 0


assert_interfaces_valid(QuadGridItem)


class Histogram2DItem(BaseImageItem):
    """2D histogram item

    Args:
        X: X coordinates (1-D array)
        Y: Y coordinates (1-D array)
        param: Histogram parameters
    """

    __implements__ = (IBasePlotItem, IBaseImageItem, IHistDataSource, IVoiImageItemType)
    _icon_name = "histogram2d.png"

    def __init__(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        param: Histogram2DParam | None = None,
        Z=None,
    ) -> None:
        if param is None:
            param = ImageParam(_("Image"))
        self._z = Z  # allows set_bins to
        super().__init__(param=param)

        # Set by parameters
        self.nx_bins = 0
        self.ny_bins = 0
        self.logscale = None

        # internal use
        self._x = None
        self._y = None

        # Histogram parameters
        self.histparam = param
        self.histparam.update_histogram(self)

        self.set_lut_range([0, 10.0])
        self.set_data(X, Y, Z)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> Histogram2DParam:
        """Return instance of the default image param DataSet"""
        return Histogram2DParam(_("2D Histogram"))

    # ---- Public API -----------------------------------------------------------
    def set_bins(self, NX: int, NY: int) -> None:
        """Set histogram bins

        Args:
            NX: Number of bins along X
            NY: Number of bins along Y
        """
        self.nx_bins = NX
        self.ny_bins = NY
        self.data = np.zeros((self.ny_bins, self.nx_bins), float)
        if self._z is not None:
            self.data_tmp = np.zeros((self.ny_bins, self.nx_bins), float)

    def set_data(
        self, X: np.ndarray, Y: np.ndarray, Z: np.ndarray | None = None
    ) -> None:
        """Set histogram data

        Args:
            X: X coordinates
            Y: Y coordinates
            Z: Z coordinates (Default value = None)
        """
        self._x = X
        self._y = Y
        self._z = Z
        self.bounds = QC.QRectF(
            QC.QPointF(X.min(), Y.min()), QC.QPointF(X.max(), Y.max())
        )
        self.update_border()

    # ---- QwtPlotItem API ------------------------------------------------------
    fill_canvas = True

    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        if self.warn_if_non_linear_scale(painter, canvasRect):
            return
        computation = self.histparam.computation
        i1, j1, i2, j2 = src_rect

        if computation == -1 or self._z is None:
            self.data[:, :] = 0.0
            nmax = histogram2d(
                self._x, self._y, i1, i2, j1, j2, self.data, self.logscale
            )
        else:
            self.data_tmp[:, :] = 0.0
            if computation in (2, 4):  # sum, avg
                self.data[:, :] = 0.0
            elif computation in (1, 5):  # min, argmin
                val = np.inf
                self.data[:, :] = val
            elif computation in (0, 6):  # max, argmax
                val = -np.inf
                self.data[:, :] = val
            elif computation == 3:
                self.data[:, :] = 1.0
            histogram2d_func(
                self._x,
                self._y,
                self._z,
                i1,
                i2,
                j1,
                j2,
                self.data_tmp,
                self.data,
                computation,
            )
            if computation in (0, 1, 5, 6):
                self.data[self.data == val] = np.nan
            else:
                self.data[self.data_tmp == 0.0] = np.nan
        if self.histparam.auto_lut:
            nmin, nmax = get_nan_range(self.data)
            self.set_lut_range([nmin, nmax])
            self.plot().update_colormap_axis(self)
        src_rect = (0, 0, self.nx_bins, self.ny_bins)

        def drawfunc(*args):
            return BaseImageItem.draw_image(self, *args)

        if self.fill_canvas:
            x1, y1, x2, y2 = canvasRect.toAlignedRect().getCoords()
            drawfunc(painter, canvasRect, src_rect, (x1, y1, x2, y2), xMap, yMap)
        else:
            dst_rect = tuple([int(i) for i in dst_rect])
            drawfunc(painter, canvasRect, src_rect, dst_rect, xMap, yMap)

    # ---- IBasePlotItem API ---------------------------------------------------
    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (
            IColormapImageItemType,
            IImageItemType,
            ITrackableItemType,
            IVoiImageItemType,
            IColormapImageItemType,
            ICSImageItemType,
        )

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        BaseImageItem.update_item_parameters(self)
        self.histparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        BaseImageItem.get_item_parameters(self, itemparams)
        itemparams.add("Histogram2DParam", self, self.histparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(
            self.histparam, itemparams.get("Histogram2DParam"), visible_only=True
        )
        self.histparam = itemparams.get("Histogram2DParam")
        self.histparam.update_histogram(self)
        BaseImageItem.set_item_parameters(self, itemparams)

    # ---- IBaseImageItem API --------------------------------------------------
    def can_sethistogram(self) -> bool:
        """
        Returns True if this item can be associated with a levels histogram

        Returns:
            bool: True if item can be associated with a levels histogram,
             False otherwise
        """
        return True

    def get_histogram(
        self, nbins: int, drange: tuple[float, float] | None = None
    ) -> tuple[np.ndarray, np.ndarray]:
        """
        Return a tuple (hist, bins) where hist is a list of histogram values

        Args:
            nbins: number of bins
            drange: lower and upper range of the bins. If not provided, range is
             simply (data.min(), data.max()). Values outside the range are ignored.

        Returns:
            Tuple (hist, bins)
        """
        if self.data is None:
            return [0], [0, 1]
        _min, _max = get_nan_range(self.data)
        if drange is not None:
            bins = np.linspace(drange[0], drange[1], nbins + 1)
        elif self.data.dtype in (np.float64, np.float32):
            bins = np.unique(
                np.array(np.linspace(_min, _max, nbins + 1), dtype=self.data.dtype)
            )
        else:
            bins = np.arange(_min, _max + 2, dtype=self.data.dtype)
        res2 = np.zeros((bins.size + 1,), np.uint32)
        _histogram(self.data.flatten(), bins, res2)
        # toc("histo2")
        res = res2[1:-1], bins
        return res


assert_interfaces_valid(Histogram2DItem)


def assemble_imageitems(
    items: list[BaseImageItem],
    src_qrect: QC.QRectF,
    destw: int,
    desth: int,
    align: int | None = None,
    add_images: bool = False,
    apply_lut: bool = False,
    apply_interpolation: bool = False,
    original_resolution: bool = False,
    force_interp_mode: str | None = None,
    force_interp_size: int | None = None,
) -> np.ndarray:
    """Assemble together image items and return resulting pixel data

    Args:
        items: List of image items
        src_qrect: Source rectangle
        destw: Destination width
        desth: Destination height
        align: Alignment (Default value = None)
        add_images: Add images (Default value = False)
        apply_lut: Apply LUT (Default value = False)
        apply_interpolation: Apply interpolation (Default value = False)
        original_resolution: Original resolution (Default value = False)
        force_interp_mode: Force interpolation mode (Default value = None)
        force_interp_size: Force interpolation size (Default value = None)

    Returns:
        Pixel data
    """
    # align width to 'align' bytes
    if align is not None:
        print(
            "plotpy.items.image.assemble_imageitems: since v2.2, "
            "the `align` option is ignored",
            file=sys.stderr,
        )
    align = 1  # XXX: Byte alignment is disabled until further notice!
    aligned_destw = int(align * ((int(destw) + align - 1) / align))
    aligned_desth = int(desth * aligned_destw / destw)

    try:
        output = np.zeros((aligned_desth, aligned_destw), np.float32)
    except ValueError:
        raise MemoryError
    if not add_images:
        dst_image = output

    dst_rect = (0, 0, aligned_destw, aligned_desth)

    src_rect = list(src_qrect.getCoords())
    # The source QRect is generally coming from a rectangle shape which is
    # adjusted to fit a given ROI on the image. So the rectangular area is
    # aligned with image pixel edges: to avoid any rounding error, we reduce
    # the rectangle area size by one half of a pixel, so that the area is now
    # aligned with the center of image pixels.
    pixel_width = src_qrect.width() / float(destw)
    pixel_height = src_qrect.height() / float(desth)
    src_rect[0] += 0.5 * pixel_width
    src_rect[1] += 0.5 * pixel_height
    src_rect[2] -= 0.5 * pixel_width
    src_rect[3] -= 0.5 * pixel_height
    # Convert to tuple for compatibility with C extension
    src_rect = tuple(src_rect)

    for it in sorted(items, key=lambda obj: obj.z()):
        if it.isVisible() and src_qrect.intersects(it.boundingRect()):
            if add_images:
                dst_image = np.zeros_like(output)
            it.export_roi(
                src_rect=src_rect,
                dst_rect=dst_rect,
                dst_image=dst_image,
                apply_lut=apply_lut,
                apply_interpolation=apply_interpolation,
                original_resolution=original_resolution,
                force_interp_mode=force_interp_mode,
                force_interp_size=force_interp_size,
            )
            if add_images:
                output += dst_image
    return output


def get_plot_qrect(plot: qwt.plot.QwtPlot, p0: QPointF, p1: QPointF) -> QRectF:
    """Get plot rectangle, in plot coordinates, from two canvas points

    Args:
        plot: Plot
        p0: First point
        p1: Second point

    Returns:
        Plot rectangle
    """
    ax, ay = X_BOTTOM, Y_LEFT
    p0x, p0y = plot.invTransform(ax, p0.x()), plot.invTransform(ay, p0.y())
    p1x, p1y = plot.invTransform(ax, p1.x() + 1), plot.invTransform(ay, p1.y() + 1)
    return QC.QRectF(p0x, p0y, p1x - p0x, p1y - p0y)


def get_items_in_rectangle(
    plot: BasePlot,
    p0: QPointF,
    p1: QPointF,
    item_type: type | None = None,
    intersect: bool = True,
) -> list[BaseImageItem]:
    """Return items which bounding rectangle intersects (p0, p1)

    Args:
        plot: Plot
        p0: First point
        p1: Second point
        item_type: Item type (if None, use `IExportROIImageItemType`)
        intersect: Intersect (Default value = True)

    Returns:
        List of items
    """
    if item_type is None:
        item_type = IExportROIImageItemType
    items = plot.get_items(item_type=item_type)
    src_qrect = get_plot_qrect(plot, p0, p1)
    if intersect:
        return [it for it in items if src_qrect.intersects(it.boundingRect())]
    else:  # contains
        return [it for it in items if src_qrect.contains(it.boundingRect())]


def compute_trimageitems_original_size(
    items: list[TrImageItem],
    src_w: list[float, float, float, float],
    src_h: list[float, float, float, float],
) -> tuple[float, float]:
    """Compute `TrImageItem` original size from max dx and dy

    Args:
        items: List of image items
        src_w: Source width
        src_h: Source height

    Returns:
        Tuple of original size
    """
    trparams = [item.get_transform() for item in items if isinstance(item, TrImageItem)]
    if trparams:
        dx_max = max([dx for _x, _y, _angle, dx, _dy, _hf, _vf in trparams])
        dy_max = max([dy for _x, _y, _angle, _dx, dy, _hf, _vf in trparams])
        return src_w / dx_max, src_h / dy_max
    return src_w, src_h


def get_image_from_qrect(
    plot: BasePlot,
    p0: QPointF,
    p1: QPointF,
    src_size: tuple[float, float] | None = None,
    adjust_range: str | None = None,
    item_type: type | None = None,
    apply_lut: bool = False,
    apply_interpolation: bool = False,
    original_resolution: bool = False,
    add_images: bool = False,
    force_interp_mode: str | None = None,
    force_interp_size: int | None = None,
) -> np.ndarray:
    """Get image pixel data from rectangle area

    Args:
        plot: Plot
        p0: First point (top-left)
        p1: Second point (bottom-right)
        src_size: Source size (Default value = None)
        adjust_range: Adjust range (Default value = None)
        item_type: Item type (Default value = None)
        apply_lut: Apply LUT (Default value = False)
        apply_interpolation: Apply interpolation (Default value = False)
        original_resolution: Original resolution (Default value = False)
        add_images: Add images (Default value = False)
        force_interp_mode: Force interpolation mode (Default value = None)
        force_interp_size: Force interpolation size (Default value = None)

    Returns:
        Image pixel data
    """
    assert adjust_range in (None, "normalize", "original")
    items = get_items_in_rectangle(plot, p0, p1, item_type=item_type)
    if not items:
        raise TypeError(_("There is no supported image item in current plot."))
    if src_size is None:
        _src_x, _src_y, src_w, src_h = get_plot_qrect(plot, p0, p1).getRect()
    else:
        # The only benefit to pass the src_size list is to avoid any
        # rounding error in the transformation computed in `get_plot_qrect`
        src_w, src_h = src_size
    destw, desth = compute_trimageitems_original_size(items, src_w, src_h)
    data = get_image_from_plot(
        plot,
        p0,
        p1,
        destw=destw,
        desth=desth,
        apply_lut=apply_lut,
        add_images=add_images,
        apply_interpolation=apply_interpolation,
        original_resolution=original_resolution,
        force_interp_mode=force_interp_mode,
        force_interp_size=force_interp_size,
    )
    if adjust_range is None:
        return data
    dtype = None
    for item in items:
        if dtype is None or item.data.dtype.itemsize > dtype.itemsize:
            dtype = item.data.dtype
    if adjust_range == "normalize":
        data = io.scale_data_to_dtype(data, dtype=dtype)
    else:
        data = np.array(data, dtype=dtype)
    return data


def get_image_in_shape(
    obj: RectangleShape,
    norm_range: bool = False,
    item_type: type | None = None,
    apply_lut: bool = False,
    apply_interpolation: bool = False,
) -> np.ndarray:
    """Get image pixel data from rectangle shape

    Args:
        obj: Rectangle shape
        norm_range: Normalize range (Default value = False)
        item_type: Item type (Default value = None)
        apply_lut: Apply LUT (Default value = False)
        apply_interpolation: Apply interpolation (Default value = False)

    Returns:
        Image pixel data
    """
    x0, y0, x1, y1 = obj.get_rect()
    (x0, x1), (y0, y1) = sorted([x0, x1]), sorted([y0, y1])
    xc0, yc0 = axes_to_canvas(obj, x0, y0)
    xc1, yc1 = axes_to_canvas(obj, x1, y1)
    adjust_range = "normalize" if norm_range else "original"
    return get_image_from_qrect(
        obj.plot(),
        QC.QPointF(xc0, yc0),
        QC.QPointF(xc1, yc1),
        src_size=(x1 - x0, y1 - y0),
        adjust_range=adjust_range,
        item_type=item_type,
        apply_lut=apply_lut,
        apply_interpolation=apply_interpolation,
        original_resolution=True,
    )


def get_image_from_plot(
    plot: BasePlot,
    p0: QPointF,
    p1: QPointF,
    destw: int | None = None,
    desth: int | None = None,
    add_images: bool = False,
    apply_lut: bool = False,
    apply_interpolation: bool = False,
    original_resolution: bool = False,
    force_interp_mode: str | None = None,
    force_interp_size: int | None = None,
) -> numpy.ndarray:
    """Get image pixel data from plot area

    Args:
        plot: Plot
        p0: First point (top-left)
        p1: Second point (bottom-right)
        destw: Destination width (Default value = None)
        desth: Destination height (Default value = None)
        add_images: Add superimposed images (instead of replace by the foreground)
        apply_lut: Apply LUT (Default value = False)
        apply_interpolation: Apply interpolation (Default value = False)
        original_resolution: Original resolution (Default value = False)
        force_interp_mode: Force interpolation mode (Default value = None)
        force_interp_size: Force interpolation size (Default value = None)

    Returns:
        Image pixel data
    """
    if destw is None:
        destw = p1.x() - p0.x() + 1
    if desth is None:
        desth = p1.y() - p0.y() + 1
    items = plot.get_items(item_type=IExportROIImageItemType)
    qrect = get_plot_qrect(plot, p0, p1)
    return assemble_imageitems(
        items,
        qrect,
        destw,
        desth,  # align=4,
        add_images=add_images,
        apply_lut=apply_lut,
        apply_interpolation=apply_interpolation,
        original_resolution=original_resolution,
        force_interp_mode=force_interp_mode,
        force_interp_size=force_interp_size,
    )



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
DownSampleCurveTool test

This plotpy tool provides a toggle to downsample the current curve with a given factor.
"""

# guitest: show

from __future__ import annotations

from guidata.qthelpers import exec_dialog, qt_app_context
from numpy import linspace, sin

from plotpy.builder import make
from plotpy.config import _


def edit_downsampled_curve(
    cdata: tuple[tuple[float, float], ...], dsamp_factor: int
) -> None:
    """
    Plot curves and return selected point(s) coordinates

    Args:
        cdata: tuple of curves to plot
        dsamp_factor: downsampling factor
    """
    win = make.dialog(
        wintitle=_("Right-click on the curve to enable/disable downsampling"),
        edit=True,
        type="curve",
        size=(800, 600),
    )
    plot = win.manager.get_plot()
    for cx, cy in cdata[:-1]:
        item = make.mcurve(cx, cy)
        plot.add_item(item)
    item = make.mcurve(*cdata[-1], "r-+", dsamp_factor=dsamp_factor, use_dsamp=True)
    plot.add_item(item)
    plot.set_active_item(item)
    plot.unselect_item(item)
    exec_dialog(win)


def test_edit_curve():
    """Test"""
    with qt_app_context(exec_loop=False):
        nlines = 1000
        x = linspace(-10, 10, num=nlines)
        y = 0.25 * sin(sin(sin(x * 0.5)))
        x2 = linspace(-10, 10, num=nlines)
        y2 = sin(sin(sin(x2)))
        cdata = ((x, y), (x2, y2), (x, sin(2 * y)))
        for dsamp_factor in (10, 50):
            edit_downsampled_curve(cdata, dsamp_factor)


if __name__ == "__main__":
    test_edit_curve()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Histogram test"""

# guitest: show

from guidata.qthelpers import qt_app_context
from numpy.random import normal

from plotpy.builder import make


def test_histogram():
    """Test"""
    with qt_app_context(exec_loop=True):
        data = normal(0, 1, (2000,))
        win = make.dialog(
            edit=False,
            toolbar=True,
            wintitle="Histogram test",
            type="curve",
        )
        plot = win.manager.get_plot()
        plot.add_item(make.histogram(data))
        win.show()


if __name__ == "__main__":
    test_histogram()



# -*- coding: utf-8 -*-

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any, TypeVar

from guidata.configtools import get_icon
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.constants import SHAPE_Z_OFFSET
from plotpy.events import (
    RectangularSelectionHandler,
    StatefulEventFilter,
    setup_standard_tool_filter,
)
from plotpy.items.shape.rectangle import RectangleShape

if TYPE_CHECKING:
    from typing import Callable

    from qtpy.QtCore import QPointF

    from plotpy.interfaces import ICurveItemType, IImageItemType
    from plotpy.items import BaseImageItem, CurveItem
    from plotpy.plot import BasePlot, PlotManager


class DefaultToolbarID:
    """Default toolbar ID"""

    pass


GuiToolT = TypeVar("GuiToolT", bound="GuiTool")


class GuiTool(QC.QObject):
    """Base class for interactive tool applying on a plot

    Args:
        manager: plot manager
        toolbar_id: toolbar ID
    """

    def __init__(
        self,
        manager: PlotManager,
        toolbar_id: Any | type[DefaultToolbarID] | None = DefaultToolbarID,
    ) -> None:
        """Constructor"""
        super().__init__()
        self.manager = manager
        self.parent_tool: GuiTool | None = None
        self.plots: set[BasePlot] = set()

        # pylint: disable=assignment-from-none
        self.action = self.create_action(manager)
        self.menu = self.create_action_menu(manager)

        if self.menu is not None:
            self.action.setMenu(self.menu)
        if toolbar_id is DefaultToolbarID:
            toolbar = manager.get_default_toolbar()
        else:
            toolbar = manager.get_toolbar(toolbar_id)
        if toolbar is not None:
            self.setup_toolbar(toolbar)

    def create_action(self, manager: PlotManager) -> QW.QAction | None:
        """Create and return tool's action

        Args:
            manager: plot manager

        Returns:
            Tool's action or None (if tool has no action)
        """
        return None

    def setup_toolbar(self, toolbar: QW.QToolBar) -> None:
        """Setup tool's toolbar

        Args:
            toolbar: toolbar
        """
        toolbar.addAction(self.action)
        if self.menu is not None:
            widget = toolbar.widgetForAction(self.action)
            widget.setPopupMode(QW.QToolButton.ToolButtonPopupMode.InstantPopup)

    def create_action_menu(self, manager: PlotManager) -> QW.QMenu | None:
        """Create and return menu for the tool's action

        Args:
            manager: plot manager
        """
        return None

    def set_parent_tool(self, tool: GuiToolT) -> None:
        """Used to organize tools automatically in menu items

        Args:
            tool: parent tool
        """
        self.parent_tool = tool

    def register_plot(self, baseplot: BasePlot) -> None:
        """Every BasePlot using this tool should call register_plot
        to notify the tool about this widget using it

        Args:
            baseplot: base plot
        """
        self.plots.add(baseplot)

    def get_active_plot(self) -> BasePlot | None:
        """Return the currently active BasePlot or None if no plot is active"""
        for plot in self.plots:
            canvas = plot.canvas()
            if canvas.hasFocus():
                return plot
        if len(self.plots) == 1:
            return list(self.plots)[0]
        return None

    def update_status(self, plot: BasePlot) -> None:
        """Update the status of the tool based on the currently active plot

        Can also be called after an action modifying the BasePlot
        (e.g. in order to update action states when an item is deselected)

        Args:
            plot: base plot
        """
        pass

    def setup_context_menu(self, menu: QW.QMenu, plot: BasePlot) -> None:
        """If the tool supports it, this method should install an action
        in the context menu

        Args:
            menu: context menu
            plot: base plot
        """
        pass


class InteractiveTool(GuiTool):
    """Interactive tool base class

    Args:
        manager: plot manager
        toolbar_id: toolbar ID
        title: tool title
        icon: tool icon
        tip: tool tip
        switch_to_default_tool: switch to default tool when finished
    """

    TITLE = None
    ICON = None
    TIP = None
    CURSOR = QC.Qt.CursorShape.CrossCursor
    SWITCH_TO_DEFAULT_TOOL = False  # switch to default tool when finished

    #: Signal emitted by InteractiveTool when validating tool action
    SIG_VALIDATE_TOOL = QC.Signal(object)

    #: Signal emitted by InteractiveTool when tool job is finished
    SIG_TOOL_JOB_FINISHED = QC.Signal()

    def __init__(
        self,
        manager: PlotManager,
        toolbar_id=DefaultToolbarID,
        title: str | None = None,
        icon: str | None = None,
        tip: str | None = None,
        switch_to_default_tool: bool | None = None,
    ):
        if title is not None:
            self.TITLE = title
        if icon is not None:
            self.ICON = icon
        if tip is not None:
            self.TIP = tip
        super().__init__(manager, toolbar_id)
        # Starting state for every plotwidget we can act upon
        self.start_state = {}

        if switch_to_default_tool is None:
            switch_to_default_tool = self.SWITCH_TO_DEFAULT_TOOL
        if switch_to_default_tool:
            self.SIG_TOOL_JOB_FINISHED.connect(self.manager.activate_default_tool)

    def create_action(self, manager: PlotManager) -> QW.QAction:
        """Create and return tool's action

        Args:
            manager: plot manager

        Returns:
            Tool's action
        """
        action = manager.create_action(
            self.TITLE, icon=get_icon(self.ICON), tip=self.TIP, triggered=self.activate
        )
        action.setCheckable(True)
        group = self.manager.get_tool_group("interactive")
        group.addAction(action)
        group.triggered.connect(self.interactive_triggered)
        return action

    def cursor(self) -> QC.Qt.CursorShape:
        """Return tool mouse cursor shape"""
        return self.CURSOR

    def register_plot(self, baseplot: BasePlot) -> None:
        """Register plot

        Args:
            baseplot: base plot
        """
        # TODO: With the introduction of PlotManager it should
        # be possible to remove the per tool dictionary start_state
        # since all plots from a manager share the same set of tools
        # the State Machine generated by the calls to tool.setup_filter
        # should be the same for all plots. Thus it should be done only once
        # and not once per plot managed by the plot manager
        super().register_plot(baseplot)
        filter = baseplot.filter
        start_state = self.setup_filter(baseplot)
        self.start_state[baseplot] = start_state
        curs = self.cursor()
        if curs is not None:
            filter.set_cursor(curs, start_state)

    def interactive_triggered(self, action: QW.QAction) -> None:
        """Slot called when the interactive tool action group is triggered.
        The purpose is to deactivate all other tools in the group.

        Note that the tool itself has already been activated because the action
        triggered the `activate` method.

        Args:
            action: tool action
        """
        if action is not self.action:
            self.deactivate()

    def activate(self) -> None:
        """Activate tool"""
        for baseplot, start_state in list(self.start_state.items()):
            baseplot.filter.set_state(start_state, None)
        self.action.setChecked(True)
        self.manager.set_active_tool(self)

    def deactivate(self) -> None:
        """Deactivate tool"""
        self.action.setChecked(False)

    def validate(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """Validate tool action

        Args:
            filter: event filter
            event: event
        """
        self.SIG_VALIDATE_TOOL.emit(filter)
        self.SIG_TOOL_JOB_FINISHED.emit()


class CommandTool(GuiTool):
    """Base class for command tools: action, context menu entry

    Args:
        manager: plot manager
        title: tool title
        icon: tool icon
        tip: tool tip
        toolbar_id: toolbar ID
    """

    CHECKABLE = False

    def __init__(
        self,
        manager: PlotManager,
        title: str,
        icon: str | None = None,
        tip: str | None = None,
        toolbar_id: Any | type[DefaultToolbarID] | None = DefaultToolbarID,
    ) -> None:
        self.title = title
        if icon and isinstance(icon, str):
            self.icon = get_icon(icon)
        else:
            self.icon = icon
        self.tip = tip
        super().__init__(manager, toolbar_id)

    def create_action(self, manager: PlotManager) -> QW.QAction:
        """Create and return tool's action

        Args:
            manager: plot manager

        Returns:
            Tool's action
        """
        return manager.create_action(
            self.title,
            icon=self.icon,
            tip=self.tip,
            triggered=self.activate,
            checkable=self.CHECKABLE,
        )

    def setup_context_menu(self, menu: QW.QMenu, plot: BasePlot) -> None:
        """Setup context menu

        Args:
            menu: context menu
            plot: base plot
        """
        menu.addAction(self.action)

    def activate(self, checked: bool = True) -> None:
        """Activate tool

        Args:
            checked: checked
        """
        plot = self.get_active_plot()
        if plot is not None:
            self.activate_command(plot, checked)

    def activate_command(self, plot: BasePlot, checked: bool) -> None:
        """Activate tool command

        Args:
            plot: base plot
            checked: checked
        """
        pass

    def set_status_active_item(self, plot: BasePlot) -> None:
        """Set status active item

        Args:
            plot: base plot
        """
        item = plot.get_active_item()
        if item:
            self.action.setEnabled(True)
        else:
            self.action.setEnabled(False)


class ActionTool(CommandTool):
    """Tool that simply associate an action to a tool

    Args:
        manager: plot manager
        action: action
        item_types: item types
        toolbar_id: toolbar ID
    """

    def __init__(
        self,
        manager: PlotManager,
        action: QW.QAction,
        item_types: Any | None = None,
        toolbar_id=DefaultToolbarID,
    ) -> None:
        self.associated_action = action
        self.item_types = item_types
        super().__init__(
            manager,
            action.text(),
            action.icon(),
            action.toolTip(),
            toolbar_id=toolbar_id,
        )

    def update_status(self, plot: BasePlot) -> None:
        """Update the status of the tool based on the currently active plot

        Can also be called after an action modifying the BasePlot
        (e.g. in order to update action states when an item is deselected)

        Args:
            plot: base plot
        """
        if self.item_types is None:
            self.action.setEnabled(True)
        else:
            items = plot.get_selected_items()
            self.action.setEnabled(
                any(isinstance(item, self.item_types) for item in items)
            )

    def create_action(self, manager: PlotManager) -> QW.QAction:
        """Create and return tool's action

        Args:
            manager: plot manager

        Returns:
            Tool's action
        """
        return self.associated_action


class ToggleTool(CommandTool):
    """Toggle tool base class

    Args:
        manager: plot manager
        title: tool title
        icon: tool icon
        tip: tool tip
        toolbar_id: toolbar ID
    """

    CHECKABLE = True

    def __init__(
        self,
        manager: PlotManager,
        title: str,
        icon: str | None = None,
        tip: str | None = None,
        toolbar_id: Any | type[DefaultToolbarID] | None = None,
    ) -> None:
        super().__init__(manager, title, icon, tip, toolbar_id)


class PanelTool(ToggleTool):
    """Panel tool base class

    Args:
        manager: plot manager
    """

    panel_id = None
    panel_name = None

    def __init__(self, manager: PlotManager) -> None:
        super().__init__(manager, self.panel_name)
        manager.get_panel(self.panel_id).SIG_VISIBILITY_CHANGED.connect(
            self.action.setChecked
        )

    def activate_command(self, plot: BasePlot, checked: bool) -> None:
        """Activate tool command

        Args:
            plot: base plot
            checked: checked
        """
        panel = self.manager.get_panel(self.panel_id)
        panel.setVisible(checked)

    def update_status(self, plot: BasePlot) -> None:
        """Update the status of the tool based on the currently active plot

        Can also be called after an action modifying the BasePlot
        (e.g. in order to update action states when an item is deselected)

        Args:
            plot: base plot
        """
        panel = self.manager.get_panel(self.panel_id)
        self.action.setChecked(panel.isVisible())


class RectangularActionTool(InteractiveTool):
    """Rectangular action tool base class

    Args:
        manager: plot manager
        func: function
        shape_style: shape style
        toolbar_id: toolbar ID
        title: tool title
        icon: tool icon
        tip: tool tip
        fix_orientation: fix orientation
        switch_to_default_tool: switch to default tool
    """

    SHAPE_STYLE_SECT = "plot"
    SHAPE_STYLE_KEY = "shape/drag"
    AVOID_NULL_SHAPE = False

    def __init__(
        self,
        manager: PlotManager,
        func: Callable,
        shape_style: tuple[str, str] | None = None,
        toolbar_id: Any | type[DefaultToolbarID] | None = DefaultToolbarID,
        title: str | None = None,
        icon: str | None = None,
        tip: str | None = None,
        fix_orientation: bool = False,
        switch_to_default_tool: bool | None = None,
    ):
        self.action_func = func
        self.fix_orientation = fix_orientation
        super().__init__(
            manager,
            toolbar_id,
            title=title,
            icon=icon,
            tip=tip,
            switch_to_default_tool=switch_to_default_tool,
        )
        if shape_style is not None:
            self.shape_style_sect = shape_style[0]
            self.shape_style_key = shape_style[1]
        else:
            self.shape_style_sect = self.SHAPE_STYLE_SECT
            self.shape_style_key = self.SHAPE_STYLE_KEY
        self.last_final_shape: RectangleShape | None = None
        self.switch_to_default_tool = switch_to_default_tool

    def get_last_final_shape(self) -> RectangleShape | None:
        """Get last final shape"""
        if self.last_final_shape is not None:
            return self.last_final_shape()

    def set_shape_style(self, shape: RectangleShape) -> None:
        """Set shape style

        Args:
            shape: shape
        """
        shape.set_style(self.shape_style_sect, self.shape_style_key)

    def create_shape(self) -> tuple[RectangleShape, int, int]:
        """Create shape

        Returns:
            A tuple with the shape, and the two handles
        """
        shape = RectangleShape(0, 0, 1, 1)
        self.set_shape_style(shape)
        return shape, 0, 2

    def setup_shape(self, shape: RectangleShape) -> None:
        """Setup shape

        Args:
            shape: shape
        """
        pass

    def get_shape(self) -> tuple[RectangleShape, int, int]:
        """Get shape

        Returns:
            A tuple with the shape, and the two handles
        """
        shape, h0, h1 = self.create_shape()
        self.setup_shape(shape)
        return shape, h0, h1

    def get_final_shape(
        self, plot: BasePlot, p0: QPointF, p1: QPointF
    ) -> RectangleShape:
        """Get final shape

        Args:
            plot: base plot
            p0: point 0
            p1: point 1

        Returns:
            Final shape
        """
        shape, h0, h1 = self.create_shape()
        self.setup_shape(shape)
        plot.add_item_with_z_offset(shape, SHAPE_Z_OFFSET)
        shape.move_local_point_to(h0, p0)
        shape.move_local_point_to(h1, p1)
        self.last_final_shape = weakref.ref(shape)
        return shape

    def get_selection_handler(
        self, filter: StatefulEventFilter, start_state: int
    ) -> RectangularSelectionHandler:
        """Get selection handler

        Args:
            filter: event filter
            start_state: start state

        Returns:
            Selection handler
        """
        return RectangularSelectionHandler(
            filter, QC.Qt.LeftButton, start_state=start_state
        )

    def setup_filter(self, baseplot: BasePlot) -> int:
        """Setup filter

        Args:
            baseplot: base plot

        Returns:
            Start state
        """
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = self.get_selection_handler(filter, start_state)
        shape, h0, h1 = self.get_shape()
        handler.set_shape(
            shape, h0, h1, self.setup_shape, avoid_null_shape=self.AVOID_NULL_SHAPE
        )
        handler.SIG_END_RECT.connect(self.end_rect)
        return setup_standard_tool_filter(filter, start_state)

    def end_rect(
        self, filter: StatefulEventFilter, p0: QC.QPointF, p1: QC.QPointF
    ) -> None:
        """End rect (this method is called when the rectangular selection is finished)

        Args:
            filter: event filter
            p0: point 0
            p1: point 1
        """
        plot = filter.plot
        if self.fix_orientation:
            left, right = min(p0.x(), p1.x()), max(p0.x(), p1.x())
            top, bottom = min(p0.y(), p1.y()), max(p0.y(), p1.y())
            self.action_func(plot, QC.QPointF(left, top), QC.QPointF(right, bottom))
        else:
            self.action_func(plot, p0, p1)
        self.SIG_TOOL_JOB_FINISHED.emit()
        if self.switch_to_default_tool:
            shape = self.get_last_final_shape()
            plot.set_active_item(shape)


class LastItemHolder:
    """Class to hold a weak reference to the last item"""

    def __init__(self, item_type: IImageItemType | ICurveItemType) -> None:
        self._item_type = item_type
        self._last_item: weakref.ReferenceType[CurveItem | BaseImageItem] | None = None

    def set(self, item: CurveItem | BaseImageItem) -> None:
        """Set the last item

        Args:
            item: BaseImageItem instance
        """
        self._last_item = weakref.ref(item)

    def get(self) -> CurveItem | BaseImageItem | None:
        """Get the last item

        Returns:
            BaseImageItem instance or None
        """
        if self._last_item is not None:
            return self._last_item()
        return None

    def update_from_selection(self, plot: BasePlot) -> CurveItem | BaseImageItem | None:
        """Update the last item from the selected items of the plot, and return it.

        Args:
            plot: BasePlot instance
        """
        items = plot.get_selected_items(item_type=self._item_type)
        if len(items) == 1:
            self.set(items[0])
        return self.get()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Flip/rotate dialog
------------------

The `FlipRotate` module provides a dialog box providing essential GUI elements
for rotating (arbitrary angle) and cropping an image:

    * :py:class:`.widgets.fliprotate.FlipRotateDialog`: dialog box
    * :py:class:`.widgets.fliprotate.FlipRotateWidget`: equivalent widget

Reference
~~~~~~~~~

.. autoclass:: FlipRotateDialog
   :members:
.. autoclass:: FlipRotateWidget
   :members:
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from guidata.configtools import get_icon
from guidata.qthelpers import create_toolbutton, win32_fix_title_bar_background
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.QtWidgets import QWidget  # Helping out python_qt_documentation

from plotpy.config import _
from plotpy.widgets import basetransform

if TYPE_CHECKING:
    from plotpy.plot import PlotOptions
    from plotpy.plot.manager import PlotManager


class FlipRotateTransform(basetransform.BaseTransform):
    """Rotate & Crop mixin class, to be mixed with a class providing the
    get_plot method, like PlotDialog or FlipRotateWidget (see below)

    Args:
        parent (QWidget): Parent widget
        manager (PlotManager): Plot manager
    """

    def __init__(self, parent: FlipRotateWidget, manager: PlotManager) -> None:
        super().__init__(parent, manager)
        self.parent = parent
        self.manager = manager

    # ------BaseTransformMixin API----------------------------------------------
    def apply_transformation(self) -> None:
        """Apply transformation, e.g. crop or rotate"""
        angle, hflip, vflip = self.parent.get_parameters()
        x, y, _a, px, py, _hf, _vf = self.item.get_transform()
        self.item.set_transform(x, y, angle * np.pi / 180, px, py, hflip, vflip)
        self.manager.get_plot().replot()

    def compute_transformation(self) -> np.ndarray:
        """Compute transformation, return compute output array

        Returns:
            Output array
        """
        angle, hflip, vflip = self.parent.get_parameters()
        data = self.item.data.copy()
        if hflip:
            data = np.fliplr(data)
        if vflip:
            data = np.flipud(data)
        if angle:
            k = int((-angle % 360) / 90)
            data = np.rot90(data, k)
        return data


class FlipRotateDialog(QW.QDialog):
    """Flip & Rotate Dialog

    Flip and rotate a :py:class:`.TrImageItem` plot item

    Args:
        parent: Parent widget
        title: Window title
        options: Options
        resize_to: Resize to (width, height)
        edit: Edit mode
        toolbar: Show toolbar
    """

    def __init__(
        self,
        parent: QWidget,
        title: str | None = None,
        options: PlotOptions | dict[str, Any] | None = None,
        resize_to: tuple[int, int] | None = None,
        edit: bool = True,
        toolbar: bool = False,
    ) -> None:
        super().__init__(parent)
        win32_fix_title_bar_background(self)

        if resize_to is not None:
            width, height = resize_to
            self.resize(width, height)

        self.button_box = None

        if title is None:
            title = _("Flip & Rotate")
        self.setWindowTitle(title)
        self.widget = FlipRotateWidget(parent=parent, options=options, toolbar=toolbar)
        self.setWindowFlags(QC.Qt.WindowType.Window)

        buttonhlayout = QW.QHBoxLayout()
        self.add_buttons_to_layout(buttonhlayout, edit)

        dialogvlayout = QW.QVBoxLayout()
        dialogvlayout.addWidget(self.widget.plot_widget.toolbar)
        dialogvlayout.addWidget(self.widget)
        dialogvlayout.addLayout(buttonhlayout)
        self.setLayout(dialogvlayout)

        self.transform = self.widget.transform
        self.plot_widget = plot_widget = self.widget.plot_widget
        self.manager = plot_widget.manager
        self.toolbar = plot_widget.toolbar

        self.accepted.connect(self.transform.accept_changes)
        self.rejected.connect(self.transform.reject_changes)

    def add_buttons_to_layout(self, layout: QW.QBoxLayout, edit: bool) -> None:
        """Add buttons to layout

        Args:
            layout: Layout
            edit: Edit mode
        """
        if edit:
            self.button_box = bbox = QW.QDialogButtonBox(
                QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel
            )
            bbox.accepted.connect(self.accept)
            bbox.rejected.connect(self.reject)
            layout.addWidget(bbox)


class FlipRotateWidget(basetransform.BaseTransformWidget):
    """Flip & Rotate Widget

    Flip and rotate a :py:class:`.TrImageItem` plot item

    Args:
        parent: Parent widget
        toolbar: Show toolbar
        options: Plot options
    """

    ROTATION_ANGLES = [str((i - 1) * 90) for i in range(4)]

    def __init__(
        self,
        parent: QWidget,
        toolbar: bool = False,
        options: PlotOptions | dict[str, Any] | None = None,
    ):
        self.angle_combo = None
        self.hflip_btn = None
        self.vflip_btn = None
        super().__init__(parent, toolbar=toolbar, options=options)
        self.transform = FlipRotateTransform(self, self.plot_widget.manager)
        self.manager = self.plot_widget.manager

    def add_buttons_to_layout(self, layout: QW.QBoxLayout) -> None:
        """Add tool buttons to layout

        Args:
            layout: Layout
        """
        # Image orientation
        angle_label = QW.QLabel(_("Angle %s:") % "(°)")
        layout.addWidget(angle_label)
        self.angle_combo = QW.QComboBox(self)
        self.angle_combo.addItems(self.ROTATION_ANGLES)
        self.angle_combo.setCurrentIndex(1)
        self.angle_combo.currentIndexChanged.connect(
            lambda index: self.transform.apply_transformation()
        )
        layout.addWidget(self.angle_combo)
        layout.addSpacing(10)

        # Image flipping
        flip_label = QW.QLabel(_("Flip:"))
        layout.addWidget(flip_label)
        hflip = create_toolbutton(
            self,
            text="",
            icon=get_icon("hflip.png"),
            toggled=lambda state: self.transform.apply_transformation(),
            autoraise=False,
        )
        self.hflip_btn = hflip
        layout.addWidget(hflip)
        vflip = create_toolbutton(
            self,
            text="",
            icon=get_icon("vflip.png"),
            toggled=lambda state: self.transform.apply_transformation(),
            autoraise=False,
        )
        self.vflip_btn = vflip
        layout.addWidget(vflip)
        layout.addSpacing(15)

        # self.add_reset_button(layout)
        basetransform.BaseTransformWidget.add_buttons_to_layout(
            self, layout, apply=False
        )

    def apply_transformation(self) -> None:
        """Apply transformation"""
        self.transform.apply_transformation()

    def reset(self) -> None:
        """Reset transformation"""
        self.angle_combo.setCurrentIndex(1)
        self.hflip_btn.setChecked(False)
        self.vflip_btn.setChecked(False)

    def set_parameters(self, angle: float, hflip: bool, vflip: bool) -> None:
        """Set transform parameters

        Args:
            angle: Angle
            hflip: Horizontal flip
            vflip: Vertical flip
        """
        angle_index = self.ROTATION_ANGLES.index(str(angle))
        self.angle_combo.setCurrentIndex(angle_index)
        self.hflip_btn.setChecked(hflip)
        self.vflip_btn.setChecked(vflip)

    def get_parameters(self) -> tuple[float, bool, bool]:
        """Return transform parameters

        Returns:
            Tuple of angle, horizontal flip, vertical flip
        """
        angle = int(str(self.angle_combo.currentText()))
        hflip = self.hflip_btn.isChecked()
        vflip = self.vflip_btn.isChecked()
        return angle, hflip, vflip


class MultipleFlipRotateWidget(basetransform.BaseMultipleTransformWidget):
    """Multiple Flip & Rotate Widget

    Flip and rotate several :py:class:`.image.TrImageItem` plot items"""

    TRANSFORM_WIDGET_CLASS = FlipRotateWidget



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""2-D Histogram test"""

# guitest: show

from guidata.qthelpers import qt_app_context
from numpy import array, concatenate, dot, random

from plotpy.builder import make
from plotpy.config import _


def hist2d_func(X, Y, Z):
    with qt_app_context(exec_loop=True):
        win = make.dialog(
            edit=True,
            toolbar=True,
            wintitle="2-D Histogram X0=(0,1), X1=(-1,-1)",
            type="image",
        )
        hist2d = make.histogram2D(X, Y, 200, 200, Z=Z, computation=2)
        curve = make.curve(
            X[::50], Y[::50], linestyle="", marker="+", title=_("Markers")
        )
        plot = win.manager.get_plot()
        plot.set_aspect_ratio(lock=False)
        plot.set_antialiasing(False)
        plot.add_item(hist2d)
        plot.add_item(curve)
        plot.set_item_visible(curve, False)
        win.show()


def hist2d(X, Y):
    with qt_app_context(exec_loop=True):
        win = make.dialog(
            edit=True,
            toolbar=True,
            wintitle="2-D Histogram X0=(0,1), X1=(-1,-1)",
            type="image",
        )
        hist2d = make.histogram2D(X, Y, 200, 200)
        curve = make.curve(
            X[::50], Y[::50], linestyle="", marker="+", title=_("Markers")
        )
        plot = win.manager.get_plot()
        plot.set_aspect_ratio(lock=False)
        plot.set_antialiasing(False)
        plot.add_item(hist2d)
        plot.add_item(curve)
        plot.set_item_visible(curve, False)
        win.show()


def test_hist_2d():
    N = 150000
    m = array([[1.0, 0.2], [-0.2, 3.0]])
    X1 = random.normal(0, 0.3, size=(N, 2))
    X2 = random.normal(0, 0.3, size=(N, 2))
    X = concatenate((X1 + [0, 1.0], dot(X2, m) + [-1, -1.0]))
    hist2d(X[:, 0], X[:, 1])


def test_hist_2d_func():
    N = 150000
    m = array([[1.0, 0.2], [-0.2, 3.0]])
    X1 = random.normal(0, 0.3, size=(N, 2))
    X2 = random.normal(0, 0.3, size=(N, 2))
    X = concatenate((X1 + [0, 1.0], dot(X2, m) + [-1, -1.0]))
    hist2d_func(X[:, 0], X[:, 1], X[:, 0] + X[:, 1])


if __name__ == "__main__":
    test_hist_2d()
    test_hist_2d_func()



# -*- coding: utf-8 -*-

from __future__ import annotations

import math
from typing import TYPE_CHECKING

import guidata.io
import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG

from plotpy.config import CONF, _
from plotpy.coords import canvas_to_axes
from plotpy.items.shape.base import AbstractShape
from plotpy.styles.shape import RangeShapeParam

if TYPE_CHECKING:
    import qwt.scale_map
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QPainter
    from qwt import QwtSymbol

    from plotpy.plot import BasePlot
    from plotpy.styles.base import ItemParameters


class BaseRangeSelection(AbstractShape):
    """Range selection shape

    Args:
        _min: Minimum value
        _max: Maximum value
        shapeparam: Shape parameters
    """

    _icon_name = ""  # This should be set in subclasses

    def __init__(
        self,
        _min: float | None = None,
        _max: float | None = None,
        shapeparam: RangeShapeParam | None = None,
    ) -> None:
        super().__init__()
        self._min = _min
        self._max = _max
        if shapeparam is None:
            self.shapeparam = RangeShapeParam(_("Range"), icon="xrange.png")
            self.shapeparam.read_config(CONF, "histogram", "range")
        else:
            self.shapeparam: RangeShapeParam = shapeparam
        self.pen = None
        self.sel_pen = None
        self.brush = None
        self.handle = None
        self.symbol = None
        self.sel_symbol = None
        if self._min is not None and self._max is not None:
            self.shapeparam.update_item(self)  # creates all the above QObjects

    def set_style(self, section: str, option: str) -> None:
        """Set style for this item

        Args:
            section: Section
            option: Option
        """
        self.shapeparam.read_config(CONF, section, option)
        self.shapeparam.update_item(self)

    def __reduce__(self) -> tuple[type, tuple, tuple]:
        """Return state information for pickling"""
        self.shapeparam.update_param(self)
        state = (self.shapeparam, self._min, self._max)
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Restore state information from pickling"""
        self.shapeparam, self._min, self._max = state
        self.shapeparam.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        self.shapeparam.update_param(self)
        writer.write(self.shapeparam, group_name="shapeparam")
        writer.write(self._min, group_name="min")
        writer.write(self._max, group_name="max")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self._min = reader.read("min")
        self._max = reader.read("max")
        self.shapeparam = RangeShapeParam(_("Range"), icon="xrange.png")
        reader.read("shapeparam", instance=self.shapeparam)
        self.shapeparam.update_item(self)

    def get_handles_pos(self) -> tuple[float, float, float]:
        """Return the handles position

        Returns:
            Tuple with three elements (x0, x1, y) or (y0, y1, x) depending on the
            orientation of the range selection. The first two elements are the
            positions of the handles, and the third element is the fixed position
            (either y or x) depending on the orientation.
        """
        raise NotImplementedError("get_handles_pos must be implemented in subclasses")

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        raise NotImplementedError("draw must be implemented in subclasses")

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        raise NotImplementedError("hit_test must be implemented in subclasses")

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        raise NotImplementedError(
            "move_local_point_to must be implemented in subclasses"
        )

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        raise NotImplementedError("move_point_to must be implemented in subclasses")

    def move_shape(
        self, old_pos: tuple[float, float], new_pos: tuple[float, float]
    ) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        raise NotImplementedError("move_shape must be implemented in subclasses")

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        raise NotImplementedError("boundingRect must be implemented in subclasses")

    def get_range(self) -> tuple[float, float]:
        """Return the range

        Returns:
            Tuple with two elements (min, max).
        """
        return self._min, self._max

    def set_range(self, _min: float, _max: float, dosignal: bool = True) -> None:
        """Set the range

        Args:
            _min: Minimum value
            _max: Maximum value
            dosignal: True to emit the SIG_RANGE_CHANGED signal
        """
        self._min = _min
        self._max = _max
        plot = self.plot()
        if dosignal and plot is not None:
            plot.SIG_RANGE_CHANGED.emit(self, self._min, self._max)

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.shapeparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        self.update_item_parameters()
        itemparams.add("ShapeParam", self, self.shapeparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.shapeparam, itemparams.get("ShapeParam"), visible_only=True)
        self.shapeparam.update_item(self)
        self.sel_brush = QG.QBrush(self.brush)


class XRangeSelection(BaseRangeSelection):
    """X range selection shape

    Args:
        _min: Minimum value
        _max: Maximum value
        shapeparam: Shape parameters
    """

    _icon_name = "xrange.png"

    def get_handles_pos(self) -> tuple[float, float, float]:
        """Return the handles position

        Returns:
            Tuple with three elements (x0, x1, y).
        """
        plot = self.plot()
        assert plot is not None, "Item must be attached to a plot"
        rct = plot.canvas().contentsRect()
        y = rct.center().y()
        x0 = plot.transform(self.xAxis(), self._min)
        x1 = plot.transform(self.xAxis(), self._max)
        return x0, x1, y

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        plot: BasePlot = self.plot()
        if not plot:
            return
        if self.selected:
            pen: QG.QPen = self.sel_pen
            sym: QwtSymbol = self.sel_symbol
        else:
            pen: QG.QPen = self.pen
            sym: QwtSymbol = self.symbol

        rct = QC.QRectF(plot.canvas().contentsRect())
        rct.setLeft(xMap.transform(self._min))
        rct.setRight(xMap.transform(self._max))

        painter.fillRect(rct, self.brush)
        painter.setPen(pen)
        painter.drawLine(rct.topLeft(), rct.bottomLeft())
        painter.drawLine(rct.topRight(), rct.bottomRight())

        dash = QG.QPen(pen)
        dash.setStyle(QC.Qt.DashLine)
        dash.setWidth(1)
        painter.setPen(dash)
        cx = rct.center().x()
        painter.drawLine(QC.QPointF(cx, rct.top()), QC.QPointF(cx, rct.bottom()))

        if self.can_resize() and not self.is_readonly():
            painter.setPen(pen)
            x0, x1, y = self.get_handles_pos()
            sym.drawSymbol(painter, QC.QPointF(x0, y))
            sym.drawSymbol(painter, QC.QPointF(x1, y))

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        x, _y = pos.x(), pos.y()
        x0, x1, _yp = self.get_handles_pos()
        d0 = math.fabs(x0 - x)
        d1 = math.fabs(x1 - x)
        d2 = math.fabs((x0 + x1) / 2 - x)
        z = np.array([d0, d1, d2])
        dist = z.min()
        handle = z.argmin()
        inside = bool(x0 < x < x1)
        return dist, handle, inside, None

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        x, _y = canvas_to_axes(self, pos)
        self.move_point_to(handle, (x, 0), ctrl)

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        val, _ = pos
        if handle == 0:
            self._min = val
        elif handle == 1:
            self._max = val
        elif handle == 2:
            move = val - (self._max + self._min) / 2
            self._min += move
            self._max += move

        self.plot().SIG_RANGE_CHANGED.emit(self, self._min, self._max)

    def move_shape(
        self, old_pos: tuple[float, float], new_pos: tuple[float, float]
    ) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        dx = new_pos[0] - old_pos[0]
        self._min += dx
        self._max += dx
        self.plot().SIG_RANGE_CHANGED.emit(self, self._min, self._max)
        self.plot().replot()

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        return QC.QRectF(self._min, 0, self._max - self._min, 0)


assert_interfaces_valid(XRangeSelection)


class YRangeSelection(BaseRangeSelection):
    """Y range selection shape

    Args:
        _min: Minimum value
        _max: Maximum value
        shapeparam: Shape parameters
    """

    _icon_name = "yrange.png"

    def get_handles_pos(self) -> tuple[float, float, float]:
        """Return the handles position

        Returns:
            Tuple with three elements (y0, y1, x).
        """
        plot = self.plot()
        assert plot is not None, "Item must be attached to a plot"
        rct = plot.canvas().contentsRect()
        x = rct.center().x()
        y0 = plot.transform(self.yAxis(), self._min)
        y1 = plot.transform(self.yAxis(), self._max)
        return y0, y1, x

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        plot: BasePlot = self.plot()
        if not plot:
            return
        if self.selected:
            pen: QG.QPen = self.sel_pen
            sym: QwtSymbol = self.sel_symbol
        else:
            pen: QG.QPen = self.pen
            sym: QwtSymbol = self.symbol

        rct = QC.QRectF(plot.canvas().contentsRect())
        rct.setTop(yMap.transform(self._max))
        rct.setBottom(yMap.transform(self._min))

        painter.fillRect(rct, self.brush)
        painter.setPen(pen)
        painter.drawLine(rct.topLeft(), rct.topRight())
        painter.drawLine(rct.bottomLeft(), rct.bottomRight())

        dash = QG.QPen(pen)
        dash.setStyle(QC.Qt.DashLine)
        dash.setWidth(1)
        painter.setPen(dash)
        cy = rct.center().y()
        painter.drawLine(QC.QPointF(rct.left(), cy), QC.QPointF(rct.right(), cy))

        if self.can_resize() and not self.is_readonly():
            painter.setPen(pen)
            y0, y1, x = self.get_handles_pos()
            sym.drawSymbol(painter, QC.QPointF(x, y0))
            sym.drawSymbol(painter, QC.QPointF(x, y1))

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        _x, y = pos.x(), pos.y()
        y0, y1, _xp = self.get_handles_pos()
        d0 = math.fabs(y0 - y)
        d1 = math.fabs(y1 - y)
        d2 = math.fabs((y0 + y1) / 2 - y)
        z = np.array([d0, d1, d2])
        dist = z.min()
        handle = z.argmin()
        inside = bool(y0 < y < y1)
        return dist, handle, inside, None

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        _x, y = canvas_to_axes(self, pos)
        self.move_point_to(handle, (0, y), ctrl)

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        _, val = pos
        if handle == 0:
            self._min = val
        elif handle == 1:
            self._max = val
        elif handle == 2:
            move = val - (self._max + self._min) / 2
            self._min += move
            self._max += move

        self.plot().SIG_RANGE_CHANGED.emit(self, self._min, self._max)

    def move_shape(
        self, old_pos: tuple[float, float], new_pos: tuple[float, float]
    ) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        dy = new_pos[1] - old_pos[1]
        self._min += dy
        self._max += dy
        self.plot().SIG_RANGE_CHANGED.emit(self, self._min, self._max)
        self.plot().replot()

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        return QC.QRectF(0, self._min, 0, self._max - self._min)


assert_interfaces_valid(YRangeSelection)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test script for locked fit parameters feature"""

import numpy as np

from plotpy.widgets.fit import FitParam, guifit


def test_locked_fit():
    """Test the curve fitting tool with locked parameters"""
    # Generate test data: y = cos(1.5*x) + 0.2
    x = np.linspace(-10, 10, 1000)
    true_offset = 0.2
    true_freq = 1.5
    y = np.cos(true_freq * x) + true_offset + np.random.rand(x.shape[0]) * 0.1

    def fit(x, params):
        """Fit function: a + cos(b*x)"""
        a, b = params
        return np.cos(b * x) + a

    # Create fit parameters
    # Lock the frequency parameter at the true value
    a = FitParam("Offset", 0.0, -1.0, 1.0)
    b = FitParam("Frequency", true_freq, 0.3, 3.0, logscale=True, locked=True)

    params = [a, b]

    print("Initial values:")
    print(f"  Offset (unlocked): {a.value}")
    print(f"  Frequency (locked): {b.value}")

    values = guifit(
        x,
        y,
        fit,
        params,
        xlabel="Time (s)",
        ylabel="Amplitude (a.u.)",
        wintitle="Locked Parameter Test",
        auto_fit=True,
    )

    if values:
        print("\nFinal values after fit:")
        print(f"  Offset: {values[0]:.4f} (should be close to {true_offset})")
        print(f"  Frequency: {values[1]:.4f} (should remain {true_freq})")
        print("\nNote: The frequency parameter was locked and should not have changed.")
        print("Only the offset parameter should have been optimized.")


if __name__ == "__main__":
    test_locked_fit()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.dataset import (
    BeginGroup,
    BeginTabGroup,
    BoolItem,
    ColorItem,
    DataSet,
    EndGroup,
    EndTabGroup,
    FloatArrayItem,
    FloatItem,
    GetAttrProp,
    ImageChoiceItem,
    IntItem,
    StringItem,
)
from qtpy import QtGui as QG
from qwt import QwtPlotMarker

from plotpy.config import _
from plotpy.styles.base import (
    MARKERSTYLE_CHOICES,
    MARKERSTYLE_NAME,
    MARKERSTYLES,
    BrushStyleItem,
    ItemParameters,
    LineStyleItem,
    SymbolItem,
    TextStyleItem,
)

if TYPE_CHECKING:
    from plotpy.items import AnnotatedShape, Axes, Marker, PolygonShape, XRangeSelection
    from plotpy.plot import BasePlot
    from plotpy.styles import (
        BrushStyleParam,
        LineStyleParam,
        SymbolParam,
        TextStyleParam,
    )


class MarkerParam(DataSet):
    """Parameters for a marker item"""

    _styles = BeginTabGroup("Styles")
    # ------------------------------------------------------------------ Line tab
    ___line = BeginGroup(_("Line")).set_prop("display", icon="dashdot.png")
    line = LineStyleItem(_("Line (not selected)"))
    sel_line = LineStyleItem(_("Line (selected)"))
    ___eline = EndGroup(_("Line"))
    # ---------------------------------------------------------------- Symbol tab
    ___sym = BeginGroup(_("Symbol")).set_prop("display", icon="diamond.png")
    symbol = SymbolItem(_("Symbol (not selected)"))
    sel_symbol = SymbolItem(_("Symbol (selected)"))
    ___esym = EndGroup(_("Symbol"))
    # ------------------------------------------------------------------ Text tab
    ___text = BeginGroup(_("Text")).set_prop("display", icon="font.png")
    text = TextStyleItem(_("Text (not selected)"))
    sel_text = TextStyleItem(_("Text (selected)"))
    ___etext = EndGroup(_("Text"))
    # ----------------------------------------------------------------------- End
    _endstyles = EndTabGroup("Styles")
    markerstyle = ImageChoiceItem(
        _("Line style"), MARKERSTYLE_CHOICES, default="NoLine"
    )
    spacing = IntItem(_("Spacing"), default=10, min=0)

    def update_param(self, obj: Marker) -> None:
        """Update parameters from object

        Args:
            obj: Marker object
        """
        self.symbol: SymbolParam
        self.text: TextStyleParam
        self.line: LineStyleParam
        self.symbol.update_param(obj.symbol())
        self.text.update_param(obj.label())
        self.line.update_param(obj.linePen())
        self.markerstyle = MARKERSTYLE_NAME[obj.lineStyle()]
        self.spacing = obj.spacing()

    def update_item(self, item: Marker) -> None:
        """Update object from parameters

        Args:
            item: Marker object
        """
        if item.selected:
            line = self.sel_line
            symb = self.sel_symbol
            text = self.sel_text
        else:
            line = self.line
            symb = self.symbol
            text = self.text
        symb.update_symbol(item)
        label = item.label()
        text.update_text(label)
        item.setLabel(label)
        item.setLinePen(line.build_pen())
        item.setLineStyle(getattr(QwtPlotMarker, self.markerstyle))
        item.setSpacing(self.spacing)
        item.update_label()

    def set_markerstyle(self, style: None | str | int) -> None:
        """Set marker line style

        Args:
            style: line style. It can be one of the following:
             * convenient values: '+', '-', '|' or None
             * `QwtPlotMarker.NoLine`, `QwtPlotMarker.Vertical`, ...
        """
        self.markerstyle = MARKERSTYLES.get(style, style)


class ShapeParam(DataSet):
    """Parameters for a shape item"""

    label = StringItem(_("Title"), default="")
    _styles = BeginTabGroup("Styles")
    # ------------------------------------------------------------------ Line tab
    ___line = BeginGroup(_("Line")).set_prop("display", icon="dashdot.png")
    line = LineStyleItem(_("Line (not selected)"))
    sel_line = LineStyleItem(_("Line (selected)"))
    ___eline = EndGroup(_("Line"))
    # ---------------------------------------------------------------- Symbol tab
    ___sym = BeginGroup(_("Symbol")).set_prop("display", icon="diamond.png")
    symbol = SymbolItem(_("Symbol (not selected)"))
    sel_symbol = SymbolItem(_("Symbol (selected)"))
    ___esym = EndGroup(_("Symbol"))
    # ------------------------------------------------------------------ Fill tab
    ___fill = BeginGroup(_("Fill pattern")).set_prop(
        "display", icon="dense6pattern.png"
    )
    fill = BrushStyleItem(_("Fill pattern (not selected)"))
    sel_fill = BrushStyleItem(_("Fill pattern (selected)"))
    ___efill = EndGroup(_("Fill pattern"))
    # ----------------------------------------------------------------------- End
    _endstyles = EndTabGroup("Styles")
    readonly = BoolItem(
        _("Read-only shape"),
        default=False,
        help=_("Read-only shapes can't be removed from the item list panel"),
    )
    private = BoolItem(
        _("Private shape"),
        default=False,
        help=_("Private shapes are not shown in the item list panel"),
    ).set_pos(col=1)

    def update_param(self, obj: PolygonShape) -> None:
        """Update parameters from object

        Args:
            obj: Shape object
        """
        self.line: LineStyleParam
        self.symbol: SymbolParam
        self.fill: BrushStyleParam
        self.sel_line: LineStyleParam
        self.sel_symbol: SymbolParam
        self.sel_fill: BrushStyleParam
        self.label = str(obj.title().text())
        self.line.update_param(obj.pen)
        self.symbol.update_param(obj.symbol)
        self.fill.update_param(obj.brush)
        self.sel_line.update_param(obj.sel_pen)
        self.sel_symbol.update_param(obj.sel_symbol)
        self.sel_fill.update_param(obj.sel_brush)
        self.readonly = obj.is_readonly()
        self.private = obj.is_private()

    def update_item(self, item: PolygonShape) -> None:
        """Update object from parameters

        Args:
            item: Shape object
        """
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.setTitle(self.label)
        item.pen = self.line.build_pen()
        item.symbol = self.symbol.build_symbol()
        item.brush = self.fill.build_brush()
        item.sel_pen = self.sel_line.build_pen()
        item.sel_symbol = self.sel_symbol.build_symbol()
        item.sel_brush = self.sel_fill.build_brush()
        item.set_readonly(self.readonly)
        item.set_private(self.private)
        if plot is not None:
            plot.blockSignals(False)


class AxesShapeParam(DataSet):
    """Parameters for an axes item"""

    arrow_angle = FloatItem(
        _("Arrow angle") + " (°)", min=0.0, max=90.0, default=15.0, nonzero=True
    )
    arrow_size = FloatItem(
        _("Arrow size") + " (%)", min=0.0, max=100.0, default=0.05, nonzero=True
    )
    _styles = BeginTabGroup("Styles")
    # ------------------------------------------------------------------ Line tab
    ___line = BeginGroup(_("Line")).set_prop("display", icon="dashdot.png")
    xarrow_pen = LineStyleItem(_("Line (X-Axis)"))
    yarrow_pen = LineStyleItem(_("Line (Y-Axis)"))
    ___eline = EndGroup(_("Line"))
    # ------------------------------------------------------------------ Fill tab
    ___fill = BeginGroup(_("Fill pattern")).set_prop(
        "display", icon="dense6pattern.png"
    )
    xarrow_brush = BrushStyleItem(_("Fill pattern (X-Axis)"))
    yarrow_brush = BrushStyleItem(_("Fill pattern (Y-Axis)"))
    ___efill = EndGroup(_("Fill pattern"))
    # ----------------------------------------------------------------------- End
    _endstyles = EndTabGroup("Styles")

    def update_param(self, item: Axes) -> None:
        """Update parameters from object

        Args:
            obj: Axes object
        """
        self.xarrow_pen: LineStyleParam
        self.yarrow_pen: LineStyleParam
        self.xarrow_brush: BrushStyleParam
        self.yarrow_brush: BrushStyleParam
        self.arrow_angle = item.arrow_angle
        self.arrow_size = item.arrow_size
        self.xarrow_pen.update_param(item.x_pen)
        self.yarrow_pen.update_param(item.y_pen)
        self.xarrow_brush.update_param(item.x_brush)
        self.yarrow_brush.update_param(item.y_brush)

    def update_item(self, item: Axes) -> None:
        """Update object from parameters

        Args:
            obj: Axes object
        """
        item.arrow_angle = self.arrow_angle
        item.arrow_size = self.arrow_size
        item.x_pen = self.xarrow_pen.build_pen()
        item.x_brush = self.xarrow_brush.build_brush()
        item.y_pen = self.yarrow_pen.build_pen()
        item.y_brush = self.yarrow_brush.build_brush()


class AnnotationParam(DataSet):
    """Parameters for annotations"""

    _multiselection = False
    show_label = BoolItem(_("Show annotation"), default=True)
    show_computations = BoolItem(
        _("Show informations on area covered by this shape"), default=True
    )
    show_subtitle = BoolItem(_("Show subtitle"), default=True)
    title = StringItem(_("Title"), default="").set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    subtitle = StringItem(_("Subtitle"), default="").set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    format = StringItem(_("String formatting"), default="%.1f")
    uncertainty = FloatItem(
        _("Uncertainty"),
        default=0.0,
        min=0.0,
        max=1.0,
        help=_("Measurement relative uncertainty"),
    ).set_pos(col=1)
    transform_matrix = FloatArrayItem(
        _("Transform matrix"), default=np.eye(3, dtype=float)
    )
    readonly = BoolItem(
        _("Read-only shape"),
        default=False,
        help=_("Read-only shapes can't be removed from the item list panel"),
    )
    private = BoolItem(
        _("Private shape"),
        default=False,
        help=_("Private shapes are not shown in the item list panel"),
    ).set_pos(col=1)

    def update_param(self, obj: AnnotatedShape) -> None:
        """Update parameters from object

        Args:
            obj: AnnotatedShape object
        """
        self.show_label = obj.is_label_visible()
        self.show_computations = obj.area_computations_visible
        self.show_subtitle = obj.subtitle_visible
        self.title = str(obj.title().text())
        self.readonly = obj.is_readonly()
        self.private = obj.is_private()

    def update_item(self, item: AnnotatedShape) -> None:
        """Update object from parameters

        Args:
            obj: AnnotatedShape object
        """
        plot: BasePlot = item.plot()
        if plot is not None:
            plot.blockSignals(True)  # Avoid unwanted calls of update_param
            # triggered by the setter methods below
        item.setTitle(self.title)
        item.set_label_visible(self.show_label)
        item.area_computations_visible = self.show_computations
        item.subtitle_visible = self.show_subtitle
        item.update_label()
        item.set_readonly(self.readonly)
        item.set_private(self.private)
        if plot is not None:
            plot.blockSignals(False)


class AnnotationParam_MS(AnnotationParam):
    """Parameters for annotations with multi-selection enabled"""

    _multiselection = True


ItemParameters.register_multiselection(AnnotationParam, AnnotationParam_MS)


class RangeShapeParam(DataSet):
    """Parameters for a range selection item"""

    label = StringItem(_("Title"), default="")
    _styles = BeginTabGroup("Styles")
    # ------------------------------------------------------------------ Line tab
    ___line = BeginGroup(_("Line")).set_prop("display", icon="dashdot.png")
    line = LineStyleItem(_("Line (not selected)"))
    sel_line = LineStyleItem(_("Line (selected)"))
    ___eline = EndGroup(_("Line"))
    # ---------------------------------------------------------------- Symbol tab
    ___symbol = BeginGroup(_("Symbol")).set_prop("display", icon="diamond.png")
    symbol = SymbolItem(_("Symbol (not selected)"))
    sel_symbol = SymbolItem(_("Symbol (selected)"))
    ___esymbol = EndGroup(_("Symbol"))
    # ------------------------------------------------------------------ Fill tab
    ___fill = BeginGroup(_("Fill")).set_prop("display", icon="dense6pattern.png")
    fill = ColorItem(_("Fill color"))
    shade = FloatItem(_("Shade"), default=0.05, min=0, max=1)
    ___efill = EndGroup(_("Fill"))
    # ----------------------------------------------------------------------- End
    _endstyles = EndTabGroup("Styles")

    def update_param(self, obj: XRangeSelection) -> None:
        """Update parameters from object

        Args:
            obj: XRangeSelection object
        """
        self.label = str(obj.title().text())
        self.line: LineStyleParam
        self.sel_line: LineStyleParam
        self.symbol: SymbolParam
        self.sel_symbol: SymbolParam
        self.line.update_param(obj.pen)
        self.sel_line.update_param(obj.sel_pen)
        self.fill = obj.brush.color().name()
        self.shade = obj.brush.color().alphaF()
        self.symbol.update_param(obj.symbol)
        self.sel_symbol.update_param(obj.sel_symbol)

    def update_item(self, item: XRangeSelection) -> None:
        """Update object from parameters

        Args:
            range: XRangeSelection object
        """
        item.setTitle(self.label)
        item.pen = self.line.build_pen()
        item.sel_pen = self.sel_line.build_pen()
        col = QG.QColor(self.fill)
        col.setAlphaF(self.shade)
        item.brush = QG.QBrush(col)
        item.symbol = self.symbol.build_symbol()
        item.sel_symbol = self.sel_symbol.build_symbol()



# cython: language_level=3
# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/__init__.py for details)

"""2D-Histogram algorithm"""

cimport cython
cimport numpy as cnp
from libc.math cimport log

cnp.import_array()

cdef inline double double_max(double a, double b): return a if a >= b else b
cdef inline double double_min(double a, double b): return a if a <= b else b

@cython.profile(False)
@cython.boundscheck(False)
def histogram2d(cnp.ndarray[double, ndim=1] X, cnp.ndarray[double, ndim=1] Y,
                double i0, double i1, double j0, double j1,
                cnp.ndarray[double, ndim=2] data, logscale):
    """Compute 2-D Histogram from data X, Y"""
    cdef double cx, cy, nmax, ix, iy
    cdef unsigned int i
    cdef Py_ssize_t n = X.shape[0]
    cdef Py_ssize_t nx = data.shape[1]
    cdef Py_ssize_t ny = data.shape[0]

    cx = nx/(i1-i0)
    cy = ny/(j1-j0)

    for i in range(n):
        #  Centered bins => - .5
        ix = (X[i] - i0) * cx - .5
        iy = (Y[i] - j0) * cy - .5
        if ix >= 0 and ix <= nx-1 and iy >= 0 and iy <= ny-1:
            data[<int> iy, <int> ix] += 1

    nmax = 0.
    if logscale:
        for j in range(ny):
            for i in range(nx):
                data[j, i] = log(1+data[j, i])
                nmax = double_max(nmax, data[j, i])
    else:
        for j in range(ny):
            for i in range(nx):
                nmax = double_max(nmax, data[j, i])
    return nmax

@cython.profile(False)
@cython.boundscheck(False)
def histogram2d_func(cnp.ndarray[double, ndim=1] X,
                     cnp.ndarray[double, ndim=1] Y,
                     cnp.ndarray[double, ndim=1] Z,
                     double i0, double i1, double j0, double j1,
                     cnp.ndarray[double, ndim=2] data_tmp,
                     cnp.ndarray[double, ndim=2] data, int computation):
    """Compute 2-D Histogram from data X, Y"""
    cdef double cx, cy, nmax, ix, iy
    cdef unsigned int i, u, v
    cdef Py_ssize_t n = X.shape[0]
    cdef Py_ssize_t nx = data.shape[1]
    cdef Py_ssize_t ny = data.shape[0]

    cx = nx/(i1-i0)
    cy = ny/(j1-j0)

    for i in range(n):
        #  Centered bins => - .5
        ix = (X[i] - i0) * cx - .5
        iy = (Y[i] - j0) * cy - .5
        if ix >= 0 and ix <= nx-1 and iy >= 0 and iy <= ny-1:
            u, v = <int> iy, <int> ix
            if computation == 0:  # max
                data_tmp[u, v] += 1
                data[u, v] = double_max(data[u, v], Z[i])
            elif computation == 1:  # min
                data_tmp[u, v] += 1
                data[u, v] = double_min(data[u, v], Z[i])
            elif computation == 2:  # sum
                data_tmp[u, v] += 1
                data[u, v] += Z[i]
            elif computation == 3:  # prod
                data_tmp[u, v] += 1
                data[u, v] *= Z[i]
            elif computation == 4:  # avg
                data_tmp[u, v] += 1
                data[u, v] += (Z[i]-data[u, v])/data_tmp[u, v]
            elif computation == 5:  # argmin
                if data[u, v] > Z[i]:
                    data_tmp[u, v] = i
                    data[u, v] = Z[i]
            elif computation == 6:  # argmax
                if data[u, v] < Z[i]:
                    data_tmp[u, v] = i
                    data[u, v] = Z[i]



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Testing image coordinates issues

Check that the first image pixel is centered on (0, 0) coordinates.

See https://github.com/PlotPyStack/guiqwt/issues/90
"""

# guitest: show

from __future__ import annotations

from typing import Literal

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv
from plotpy.tools import DisplayCoordsTool


def test_pixel_coords(image_type: Literal["standard", "xy"] = "standard") -> None:
    """Testing image pixel coordinates"""
    title = test_pixel_coords.__doc__ + f" ({image_type} image)"
    data = ptd.gen_2d_gaussian(20, np.uint8, x0=-10, y0=-10, mu=7, sigma=10.0)
    with qt_app_context(exec_loop=True):
        if image_type == "xy":
            x = np.linspace(0.0, 10.0, data.shape[1], dtype=float)
            y = np.linspace(0.0, 10.0, data.shape[0], dtype=float) ** 2 / 10.0
            image = make.xyimage(x, y, data, interpolation="nearest")
        else:
            image = make.image(data, interpolation="nearest")
        text = "First pixel should be centered on (0, 0) coordinates"
        label = make.label(text, (1.0, 1.0), (0, 0), "L")
        rect = make.rectangle(5.0, 5.0, 10.0, 10.0, "Rectangle")
        cursors = []
        for i_cursor in range(0, 21, 10):
            cursors.append(make.vcursor(i_cursor, movable=False))
            cursors.append(make.hcursor(i_cursor, movable=False))
        win = ptv.show_items([image, label, rect] + cursors, wintitle=title)
        plot = win.get_plot()
        plot.select_item(image)
        win.manager.get_tool(DisplayCoordsTool).activate_curve_pointer(True)


if __name__ == "__main__":
    test_pixel_coords("standard")
    test_pixel_coords("xy")



# -*- coding: utf-8 -*-

from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, create_action
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.config import _
from plotpy.constants import ID_LCS, ID_OCS, ID_XCS, ID_YCS
from plotpy.interfaces import IPanel
from plotpy.panels.base import PanelWidget
from plotpy.panels.csection.csplot import (
    BaseCrossSectionPlot,
    LineCrossSectionPlot,
    ObliqueCrossSectionPlot,
    XCrossSectionPlot,
    YCrossSectionPlot,
)
from plotpy.tools import ExportItemDataTool


class CrossSectionWidget(PanelWidget):
    """Base class for cross section panels

    Args:
        parent: parent widget
    """

    SIG_RESIZED = QC.Signal()  # emitted when the widget is resized
    PANEL_ID = None
    PANEL_TITLE = _("Cross section tool")
    PANEL_ICON = "csection.png"
    CrossSectionPlotKlass = BaseCrossSectionPlot  # to be overridden in subclasses

    __implements__ = (IPanel,)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.export_ac = None
        self.autoscale_ac = None
        self.refresh_ac = None
        self.autorefresh_ac = None
        self.lockscales_ac = None

        self.manager = None  # manager for the associated image plot

        # Avoid circular import
        # pylint-disable=import-outside-toplevel
        from plotpy.plot.manager import PlotManager

        self.local_manager = PlotManager(self)
        self.cs_plot = self.CrossSectionPlotKlass(parent)
        self.cs_plot.SIG_CS_CURVE_CHANGED.connect(self.cs_curve_has_changed)
        self.export_tool = None
        self.setup_plot()

        self.toolbar = QW.QToolBar(self)
        self.toolbar.setOrientation(QC.Qt.Orientation.Vertical)

        self.setup_widget()

    # Reimplement Qt method for sending a signal when the widget is resized
    def resizeEvent(self, event):
        """

        :param event:
        """
        super().resizeEvent(event)
        self.SIG_RESIZED.emit()

    def set_options(self, autoscale=None, autorefresh=None, lockscales=None):
        """

        :param autoscale:
        :param autorefresh:
        :param lockscales:
        """
        assert self.manager is not None, (
            "Panel '%s' must be registered to plot manager before changing options"
            % self.PANEL_ID
        )
        if autoscale is not None:
            self.autoscale_ac.setChecked(autoscale)
        if autorefresh is not None:
            self.autorefresh_ac.setChecked(autorefresh)
        if lockscales is not None:
            self.lockscales_ac.setChecked(lockscales)

    def setup_plot(self):
        """ """
        # Configure the local manager
        lman = self.local_manager
        lman.add_plot(self.cs_plot)
        lman.register_all_curve_tools()
        self.export_tool = lman.get_tool(ExportItemDataTool)

    def setup_widget(self):
        """ """
        layout = QW.QHBoxLayout()
        layout.addWidget(self.cs_plot)
        layout.addWidget(self.toolbar)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def cs_curve_has_changed(self, curve):
        """Cross section curve has just changed"""
        # Do something with curve's data for example
        pass

    def register_panel(self, manager):
        """Register panel to plot manager"""
        self.manager = manager
        for plot in manager.get_plots():
            self.cs_plot.connect_plot(plot)
        self.setup_actions()
        self.add_actions_to_toolbar()

    def configure_panel(self):
        """Configure panel"""
        pass

    def get_plot(self):
        """

        :return:
        """
        return self.manager.get_active_plot()

    def setup_actions(self):
        """ """
        self.export_ac = self.export_tool.action
        self.lockscales_ac = create_action(
            self,
            _("Lock scales"),
            icon=get_icon("axes.png"),
            toggled=self.cs_plot.toggle_lockscales,
            tip=_("Lock scales to main plot axes"),
        )
        self.lockscales_ac.setChecked(self.cs_plot.lockscales)
        self.autoscale_ac = create_action(
            self,
            _("Auto-scale"),
            icon=get_icon("csautoscale.png"),
            toggled=self.cs_plot.toggle_autoscale,
        )
        self.autoscale_ac.toggled.connect(self.lockscales_ac.setDisabled)
        self.autoscale_ac.setChecked(self.cs_plot.autoscale_mode)
        self.refresh_ac = create_action(
            self,
            _("Refresh"),
            icon=get_icon("refresh.png"),
            triggered=lambda: self.cs_plot.update_plot(),
        )
        self.autorefresh_ac = create_action(
            self,
            _("Auto-refresh"),
            icon=get_icon("autorefresh.png"),
            toggled=self.cs_plot.toggle_autorefresh,
        )
        self.autorefresh_ac.setChecked(self.cs_plot.autorefresh_mode)

    def add_actions_to_toolbar(self):
        """ """
        add_actions(
            self.toolbar,
            (
                self.export_ac,
                None,
                self.autoscale_ac,
                self.lockscales_ac,
                None,
                self.refresh_ac,
                self.autorefresh_ac,
            ),
        )

    def register_shape(self, shape, refresh=True):
        """

        :param shape:
        :param final:
        :param refresh:
        """
        plot = self.get_plot()
        self.cs_plot.register_shape(plot, shape, refresh)

    def unregister_shape(self, shape):
        """

        :param shape:
        """
        self.cs_plot.unregister_shape(shape)

    def update_plot(self, obj=None):
        """
        Update cross section curve(s) associated to object *obj*

        *obj* may be a marker or a rectangular shape
        (see :py:class:`.tools.CrossSectionTool`
        and :py:class:`.tools.AverageCrossSectionTool`)

        If obj is None, update the cross sections of the last active object
        """
        self.cs_plot.update_plot(obj)


assert_interfaces_valid(CrossSectionWidget)


class XCrossSection(CrossSectionWidget):
    """X-axis cross section panel

    Args:
        parent: parent widget
    """

    PANEL_ID = ID_XCS
    OTHER_PANEL_ID = ID_YCS
    CrossSectionPlotKlass = XCrossSectionPlot

    def __init__(self, parent=None):
        super().__init__(parent)
        self.peritem_ac = None
        self.applylut_ac = None

    def set_options(
        self,
        autoscale=None,
        autorefresh=None,
        peritem=None,
        applylut=None,
        lockscales=None,
    ):
        """

        :param autoscale:
        :param autorefresh:
        :param peritem:
        :param applylut:
        :param lockscales:
        """
        assert self.manager is not None, (
            f"Panel '{self.PANEL_ID}' must be "
            "registered to plot manager before changing options"
        )

        if autoscale is not None:
            self.autoscale_ac.setChecked(autoscale)
        if autorefresh is not None:
            self.autorefresh_ac.setChecked(autorefresh)
        if lockscales is not None:
            self.lockscales_ac.setChecked(lockscales)
        if peritem is not None:
            self.peritem_ac.setChecked(peritem)
        if applylut is not None:
            self.applylut_ac.setChecked(applylut)

    def add_actions_to_toolbar(self):
        """ """
        other = self.manager.get_panel(self.OTHER_PANEL_ID)
        if other is None:
            add_actions(
                self.toolbar,
                (
                    self.peritem_ac,
                    self.applylut_ac,
                    None,
                    self.export_ac,
                    None,
                    self.autoscale_ac,
                    self.lockscales_ac,
                    None,
                    self.refresh_ac,
                    self.autorefresh_ac,
                ),
            )
        else:
            add_actions(
                self.toolbar,
                (
                    other.peritem_ac,
                    other.applylut_ac,
                    None,
                    self.export_ac,
                    None,
                    other.autoscale_ac,
                    other.lockscales_ac,
                    None,
                    other.refresh_ac,
                    other.autorefresh_ac,
                ),
            )
            other.peritem_ac.toggled.connect(self.cs_plot.toggle_perimage_mode)
            other.applylut_ac.toggled.connect(self.cs_plot.toggle_apply_lut)
            other.autoscale_ac.toggled.connect(self.cs_plot.toggle_autoscale)
            other.refresh_ac.triggered.connect(lambda: self.cs_plot.update_plot())
            other.autorefresh_ac.toggled.connect(self.cs_plot.toggle_autorefresh)
            other.lockscales_ac.toggled.connect(self.cs_plot.toggle_lockscales)

    def closeEvent(self, event):
        """

        :param event:
        """
        self.hide()
        event.ignore()

    def setup_actions(self):
        """ """
        CrossSectionWidget.setup_actions(self)
        self.peritem_ac = create_action(
            self,
            _("Per image cross-section"),
            icon=get_icon("csperimage.png"),
            toggled=self.cs_plot.toggle_perimage_mode,
            tip=_(
                "Enable the per-image cross-section mode, "
                "which works directly on image rows/columns.\n"
                "That is the fastest method to compute "
                "cross-section curves but it ignores "
                "image transformations (e.g. rotation)"
            ),
        )
        self.applylut_ac = create_action(
            self,
            _("Apply LUT\n(contrast settings)"),
            icon=get_icon("csapplylut.png"),
            toggled=self.cs_plot.toggle_apply_lut,
            tip=_(
                "Apply LUT (Look-Up Table) contrast settings.\n"
                "This is the easiest way to compare images "
                "which have slightly different level ranges.\n\n"
                "Note: LUT is coded over 1024 levels (0...1023)"
            ),
        )
        self.peritem_ac.setChecked(True)
        self.applylut_ac.setChecked(False)


class YCrossSection(XCrossSection):
    """Y-axis cross section panel

    Args:
        parent: parent widget
        position: position of the panel ('left' or 'right')
        xsection_pos: position of the cross section plot ('top' or 'bottom')
    """

    PANEL_ID = ID_YCS
    OTHER_PANEL_ID = ID_XCS
    CrossSectionPlotKlass = YCrossSectionPlot

    def __init__(self, parent=None, position="right", xsection_pos="top"):
        self.xsection_pos = xsection_pos
        self.spacer = QW.QSpacerItem(0, 0)
        super().__init__(parent)
        self.cs_plot.set_axis_direction("bottom", reverse=position == "left")

    def setup_widget(self):
        """ """
        toolbar = self.toolbar
        toolbar.setOrientation(QC.Qt.Orientation.Horizontal)
        layout = QW.QVBoxLayout()
        if self.xsection_pos == "top":
            layout.addSpacerItem(self.spacer)
        layout.addWidget(toolbar)
        layout.addWidget(self.cs_plot)
        if self.xsection_pos == "bottom":
            layout.addSpacerItem(self.spacer)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def adjust_height(self, height):
        """

        :param height:
        """
        self.spacer.changeSize(0, height, QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        self.layout().invalidate()


# Oblique cross section panel
class ObliqueCrossSection(CrossSectionWidget):
    """Oblique averaged cross section panel

    Args:
        parent: parent widget
    """

    PANEL_ID = ID_OCS
    CrossSectionPlotKlass = ObliqueCrossSectionPlot
    PANEL_ICON = "csection_oblique.png"

    def setup_actions(self):
        """ """
        super().setup_actions()
        self.lockscales_ac.setChecked(False)
        self.autoscale_ac.setChecked(True)


class LineCrossSection(CrossSectionWidget):
    """Line cross section panel

    Args:
        parent: parent widget
    """

    PANEL_ID = ID_LCS
    CrossSectionPlotKlass = LineCrossSectionPlot

    def __init__(self, parent=None):
        super().__init__(parent)
        self.cs_plot.set_axis_direction("bottom", reverse=False)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test showing SVG shapes"""

# guitest: show

from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path
from plotpy.tests import vistools as ptv


def test_svg():
    """Test showing SVG shapes"""
    with qt_app_context(exec_loop=True):
        path1, path2 = get_path("svg_tool.svg"), get_path("svg_target.svg")
        csvg = make.svg("circle", path2, 0, 0, 100, 100, "Circle")
        rsvg = make.svg("rectangle", path1, 150, 0, 200, 100, "Rect")
        ssvg = make.svg("square", path1, 0, 150, 100, 250, "Square")
        items = [csvg, rsvg, ssvg]
        _win = ptv.show_items(items, wintitle=test_svg.__doc__, lock_aspect_ratio=True)


if __name__ == "__main__":
    test_svg()



#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Plot Image unit tests"""

import contextlib

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.builder import make
from plotpy.plot import BasePlot
from plotpy.tools import ColormapTool, ReverseColormapTool
from plotpy.tools.image import LockTrImageTool


def compute_image():
    """Return a test image as numpy array"""
    N = 2000
    x = np.array(np.linspace(-5, 5, N), np.float32)
    img = np.zeros((N, N), np.float32)
    x.shape = (1, N)
    img += x**2
    return img


@contextlib.contextmanager
def plot_image(item):
    win = make.dialog(wintitle="Image plot test", type="image", toolbar=True)
    plot = win.manager.get_plot()
    assert isinstance(plot, BasePlot)
    plot.add_item(item)
    assert item in plot.items
    assert item.plot() is plot
    win.show()
    yield plot
    exec_dialog(win)


def test_plot_image():
    """Test plotting of an Image"""
    with qt_app_context(exec_loop=False):
        item = make.image(compute_image())
        with plot_image(item):
            pass


def compute_image_xy():
    N = 2000
    T = np.float32
    x = np.array(np.linspace(-5, 5, N), T)
    img = np.zeros((N, N), T)
    x.shape = (1, N)
    img += x**2
    x.shape = (N, 1)
    img += x**2
    x.shape = (N,)
    return x, (x + 5) ** 0.6, img


def test_plot_image_xy():
    """Test plotting of an Image with custom XY coordinates"""
    with qt_app_context(exec_loop=False):
        item = make.xyimage(*compute_image_xy())
        with plot_image(item):
            pass


def test_plot_quad_image():
    """Test plotting of an Image with custom quad coordinates"""
    delta = 0.025
    x = np.arange(-3.0, 3.0, delta)
    y = np.arange(-2.0, 2.0, delta)
    X, Y = np.meshgrid(x, y)
    Z = X * Y
    with qt_app_context(exec_loop=False):
        item = make.quadgrid(X, Y, Z)
        with plot_image(item):
            pass


def test_plot_tr_image():
    """Test plotting of a TrImageItem"""
    with qt_app_context(exec_loop=False):
        img = compute_image()
        item = make.trimage(img)
        with plot_image(item) as plot:
            # translate the image and check the new bounds
            bounds1 = item.bounds
            item.set_transform(50, 25, 0)
            bounds2 = item.bounds
            assert bounds1.width() == bounds2.width()
            assert bounds1.height() == bounds2.height()
            assert bounds2.left() == bounds1.left() + 50
            assert bounds2.top() == bounds1.top() + 25

            # rescale the image
            item.set_transform(0, 0, 0, 2.0, 3.0)
            plot.do_autoscale()
            bounds3 = item.bounds
            assert bounds3.width() == bounds2.width() * 2.0
            assert bounds3.height() == bounds2.height() * 3.0


def test_lock_tr_image():
    """Test LockTrImageTool"""
    with qt_app_context(exec_loop=True):
        win = make.dialog(type="image", toolbar=True)
        win.show()
        item = make.trimage(compute_image())
        # with plot_image(item) as plot:
        plot = win.manager.get_plot()
        plot.add_item(item)
        plot.select_item(item)
        tool = win.manager.add_tool(LockTrImageTool)
        tool.activate()
        assert item.is_locked()
        tool.activate(False)
        assert not item.is_locked()
        exec_dialog(win)


@pytest.mark.parametrize("ratio", [1.0, 0.75, 1.5, 2.0, 3.0])
def test_set_aspect_ratio(ratio):
    """Test BasePlot.set_aspect_ratio method()

    It ensures that the new height is correctly set."""
    with qt_app_context(exec_loop=False):
        win = make.dialog(type="image")
        item = make.image(compute_image())
        plot = win.manager.get_plot()
        plot.add_item(item, autoscale=False)
        win.show()
        x0, x1, y0, y1 = plot.get_plot_limits()
        plot.set_aspect_ratio(ratio, True)
        assert plot.get_aspect_ratio() == ratio
        exec_dialog(win)


def test_colormap_tools():
    """Test ColorMapTool and ReverseColormapTool on an image"""
    with qt_app_context(exec_loop=False):
        win = make.dialog(type="image", toolbar=True)
        item = make.image(compute_image())
        plot = win.manager.get_plot()
        plot.add_item(item)
        win.show()

        # default color map should be "jet"
        color_map_tool = win.manager.get_tool(ColormapTool)
        cmap = item.get_color_map()
        assert cmap.name == "jet"
        jet_img = plot.grab().toImage()

        # change the colormap
        plot.select_item(item)
        cmap_name = "accent"
        color_map_tool.activate_cmap(cmap_name)
        cmap = item.get_color_map()
        assert cmap.name == cmap_name
        accent_img = plot.grab().toImage()
        assert jet_img != accent_img

        # reverse the colormap
        reverse_tool = win.manager.get_tool(ReverseColormapTool)
        assert not cmap.invert, "Colormap should not be inverted"
        reverse_tool.activate()
        cmap = item.get_color_map()
        assert cmap.invert, "Colormap should be inverted"

        exec_dialog(win)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
LUT range
---------

The :mod:`plotpy.lutrange` module provides functions to compute and manipulate
the LUT range of plot items.

.. autofunction:: hist_range_threshold

.. autofunction:: lut_range_threshold
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from plotpy.items import BaseImageItem, Histogram2DItem
    from plotpy.items.histogram import HistDataSource


def hist_range_threshold(
    hist: np.ndarray, bin_edges: np.ndarray, percent: float
) -> tuple[float, float]:
    """Return the range corresponding to the central `percent` of the histogram mass.
    This can be used to eliminate outliers symmetrically, e.g. for contrast adjustment.

    Notes:
        - If the histogram comes from an image with integer values (e.g. 0-255),
          the first bin is assumed to correspond to value 0 and will be ignored.
        - For floating-point images, all bins are considered.

    Args:
        hist: The histogram (length N)
        bin_edges: The bin edges (length N+1)
        percent: Percent of histogram mass to retain (between 0 and 100)

    Returns:
        A tuple containing the minimum and maximum values of the range
        corresponding to the central `percent` of the histogram mass.
    """
    if not (0 <= percent <= 100):
        raise ValueError("percent must be in (0, 100]")

    hist_len = len(hist)
    i_offset = 0

    # 1. Remove the first bin (typically corresponding to 0), only for integer images
    if np.issubdtype(bin_edges.dtype, np.integer):
        hist = hist[1:]
        i_offset = 1

    # 3. Threshold: keep `percent`% of the mass → remove (1 - percent)% symmetrically
    threshold = 0.5 * percent / 100 * hist.sum()

    # 4. Find index where left cumulative sum exceeds threshold
    i_bin_min = max(np.cumsum(hist).searchsorted(threshold) - i_offset, 0)

    # 5. Find index where right cumulative sum exceeds threshold
    i_bin_max = hist_len - np.searchsorted(np.cumsum(np.flipud(hist)), threshold)

    # 6. Return bounds as [bin_edges[i_min], bin_edges[i_max + 1]]
    vmin, vmax = bin_edges[i_bin_min], bin_edges[i_bin_max]

    return vmin, vmax


def lut_range_threshold(
    item: BaseImageItem | Histogram2DItem | HistDataSource, bins: int, percent: float
) -> tuple[float, float]:
    """Return the lut range corresponding to the given percent of the histogram

    Args:
        item (BaseImageItem | Histogram2DItem | HistDataSource): The plot item
        bins (int): The number of bins
        percent (float): The percent of the histogram

    Returns:
        tuple[float, float]: The lut range
    """
    hist, bin_edges = item.get_histogram(bins)
    return hist_range_threshold(hist, bin_edges, percent)



# content of conftest.py

import gc
import os

import guidata
import h5py
import numpy
import PIL
import pytest
import qtpy
import qwt
import scipy
import tifffile
from guidata.config import ValidationMode, set_validation_mode
from guidata.env import execenv
from guidata.utils.gitreport import format_git_info_for_pytest, get_git_info_for_modules

import plotpy

# Set validation mode to STRICT for the sigima package
set_validation_mode(ValidationMode.STRICT)

# Turn on unattended mode for executing tests without user interaction
execenv.unattended = True
execenv.verbose = "quiet"


def pytest_addoption(parser):
    """Add custom command line options to pytest."""
    parser.addoption(
        "--show-windows",
        action="store_true",
        default=False,
        help="Display Qt windows during tests (disables QT_QPA_PLATFORM=offscreen)",
    )


def pytest_configure(config):
    """Configure pytest based on command line options."""
    if config.option.durations is None:
        config.option.durations = 10  # Default to showing 10 slowest tests
    config.addinivalue_line(
        "markers",
        "requires_display: mark test as requiring a display "
        "(skipped in offscreen mode)",
    )
    if not config.getoption("--show-windows"):
        os.environ.setdefault("QT_QPA_PLATFORM", "offscreen")


def pytest_runtest_setup(item):
    """Run setup before each test."""
    if "requires_display" in item.keywords:
        if os.environ.get("QT_QPA_PLATFORM") == "offscreen":
            pytest.skip("Skipped in offscreen mode (requires display)")


@pytest.fixture(scope="session", autouse=True)
def disable_gc_for_tests():
    """Disable garbage collection for all tests in the session."""
    # Important note:
    # ---------------
    # We need to disable garbage collection for all tests in the session because
    # this test suite is not representative of a typical application.
    # The high level of stress on the garbage collector can lead to false positives
    # in tests that rely on reference counting or finalization.
    # In a typical application, the garbage collector should be left enabled.

    gc.disable()
    yield
    gc.enable()


def pytest_report_header(config):
    """Add additional information to the pytest report header."""
    qtbindings_version = qtpy.PYSIDE_VERSION
    if qtbindings_version is None:
        qtbindings_version = qtpy.PYQT_VERSION
    infolist = [
        f"PlotPy {plotpy.__version__}",
        f"  guidata {guidata.__version__}, PythonQwt {qwt.__version__}, "
        f"{qtpy.API_NAME} {qtbindings_version} [Qt version: {qtpy.QT_VERSION}]",
        f"  NumPy {numpy.__version__}, SciPy {scipy.__version__}, "
        f"h5py {h5py.__version__}, Pillow {PIL.__version__}, "
        f"tifffile {tifffile.__version__}",
    ]

    # Git information for all modules using the gitreport module
    modules_config = [
        ("PlotPy", plotpy, "."),  # PlotPy uses current directory
        ("guidata", guidata, None),
    ]
    git_repos = get_git_info_for_modules(modules_config)
    git_info_lines = format_git_info_for_pytest(git_repos, "PlotPy")
    if git_info_lines:
        infolist.extend(git_info_lines)

    return infolist



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test plotting of a curve with PlotDialog"""

import numpy as np
from guidata.qthelpers import exec_dialog, qt_app_context
from qwt import QwtPlotItem

from plotpy.builder import make
from plotpy.plot import PlotWidget
from plotpy.tools import (
    AntiAliasingTool,
    AxisScaleTool,
    CurveStatsTool,
    DownSamplingTool,
)


def test_plot_curve_anomalies():
    """Test plotting of a curve with anomalies"""
    x = np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0])
    dx = dy = y = x * np.nan

    with qt_app_context(exec_loop=False):
        win = make.dialog(toolbar=True, type="curve")
        plot = win.manager.get_plot()
        for curve in (
            make.merror(x, y, label="Curve"),
            make.merror(x, y, dx, label="Curve"),
            make.merror(x, y, dy, label="Curve"),
            make.merror(x, y, dx, dy, label="Curve"),
        ):
            plot.add_item(curve)
        win.show()
        plot.do_autoscale()  # Force autoscale (was crashing initially with NaN data)
        exec_dialog(win)


def test_plot_curve():
    """Test plotting of a curve with PlotDialog"""
    x = np.linspace(-10, 10, 200)
    y = np.sin(x)

    with qt_app_context(exec_loop=False):
        curve = make.curve(x, y, color="g", curvestyle="Sticks")
        curve.setTitle("Curve")

        win = make.dialog(toolbar=True, type="curve")
        plot = win.manager.get_plot()
        plot.add_item(curve)

        win.show()

        assert isinstance(win.plot_widget, PlotWidget)
        assert win.plot_widget.plot == plot

        # Check that specific curve tools are added
        for curve_tool_class in (
            CurveStatsTool,
            AntiAliasingTool,
            AxisScaleTool,
            DownSamplingTool,
        ):
            tool = win.manager.get_tool(curve_tool_class)
            assert tool is not None, f"Tool of type {curve_tool_class} not found"

        exec_dialog(win)


def test_plot_curve_anti_aliasing():
    """Test Anti-aliasing tool"""
    x = np.linspace(-10, 10, 200)
    y = np.sin(x)

    with qt_app_context(exec_loop=False):
        curve = make.curve(x, y, color="g")

        win = make.dialog(toolbar=True, type="curve")
        plot = win.manager.get_plot()
        plot.add_item(curve)

        win.show()

        assert isinstance(win.plot_widget, PlotWidget)
        assert win.plot_widget.plot == plot

        anti_aliasing_tool = win.manager.get_tool(AntiAliasingTool)
        assert anti_aliasing_tool is not None, "AntiAliasingTool not found"

        original_img = plot.grab().toImage()

        # Activate anti aliasing
        anti_aliasing_tool.activate(True)
        assert plot.antialiased
        assert curve.testRenderHint(QwtPlotItem.RenderAntialiased)
        antialiased_img = plot.grab().toImage()
        assert antialiased_img != original_img

        # Deactivate anti aliasing
        anti_aliasing_tool.activate(False)
        assert not plot.antialiased
        assert not curve.testRenderHint(QwtPlotItem.RenderAntialiased)
        aliased_img = plot.grab().toImage()
        assert original_img == aliased_img

        exec_dialog(win)


if __name__ == "__main__":
    test_plot_curve_anomalies()
    # test_plot_curve()
    # test_plot_curve_anti_aliasing()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Curve plotting test with large datasets"""

# guitest: show

import timeit

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def __create_curve(npts: int, linewidth: float):
    """Create a curve with random data"""
    x = np.linspace(-10, 10, npts)
    y = np.random.normal(size=npts)
    curve = make.curve(x, y, color="g", linewidth=linewidth)
    return curve


def __show_curve(curve, benchmark: bool = False):
    """Show the curve in a plot window"""
    with qt_app_context(exec_loop=not benchmark):
        _win = ptv.show_items(
            [curve],
            wintitle=test_curve_rendering_performance.__doc__,
            title="Curves",
            plot_type="curve",
            disable_readonly_for_items=False,
        )
        if benchmark:
            QW.QApplication.processEvents()
    return _win


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_interactive_curve_rendering(npts: int = 500000, linewidth: float = 1.0):
    """Interactive test for curve rendering with large datasets"""
    curve = __create_curve(npts, linewidth)
    _win = __show_curve(curve, benchmark=False)


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_curve_rendering_performance(npts: int = 500000, linewidth: float = 1.0):
    """Test curve rendering performance with large datasets"""
    curve = __create_curve(npts, linewidth)

    # Show once before benchmarking
    _win = __show_curve(curve, benchmark=True)

    # Benchmark with multiple runs
    n_runs = 5
    times = []

    print("Benchmarking in progress", end="")
    for i in range(n_runs):
        start_time = timeit.default_timer()
        _win = __show_curve(curve, benchmark=True)
        elapsed_time = timeit.default_timer() - start_time
        times.append(elapsed_time)
        print(".", end="", flush=True)
    print(" done.")

    # Display statistics
    print(
        f"Results ({n_runs} runs): {np.mean(times):.4f} ± {np.std(times):.4f} seconds"
    )


if __name__ == "__main__":
    # test_interactive_curve_rendering(10000, 2.0, False)
    # Test with different linewidths and antialiasing settings
    print("\n" + "=" * 70)
    print("Curve Rendering Performance Benchmark")
    print("=" * 70)

    print("\n--- Test 1: linewidth=1.0 ---")
    test_curve_rendering_performance(10000, 1.0)

    print("\n--- Test 2: linewidth=2.0 ---")
    test_curve_rendering_performance(10000, 2.0)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Horizontal/vertical cursors test"""

# guitest: show

from guidata.qthelpers import qt_app_context
from numpy import linspace, sin

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def test_cursor():
    """Test cursor"""
    x = linspace(-10, 10, 1000) + 1
    y = sin(sin(sin(x))) + 3
    with qt_app_context(exec_loop=True):
        curve = make.curve(x, y, "ab", "b")
        hcursor = make.hcursor(3.2, label="y = %.2f")
        vcursor = make.vcursor(7, label="x = %.2f")
        vcursor2 = make.vcursor(-1, label="NOT MOVABLE = %.2f", movable=False)
        xcursor = make.xcursor(-4, 2.5, label="x = %.2f<br>y = %.2f")
        legend = make.legend("TR")
        _win = ptv.show_items(
            wintitle="Plot cursors",
            items=[curve, hcursor, vcursor, vcursor2, xcursor, legend],
            plot_type="curve",
        )


if __name__ == "__main__":
    test_cursor()
